import sys
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Union, Optional, Literal, TypedDict

import brainrender
import numpy as np
from brainglobe_atlasapi.bg_atlas import BrainGlobeAtlas
from brainrender.actors import Points
from typing_extensions import TypeAlias

from neuralib.argp import AbstractParser, argument, str_tuple_type
from neuralib.atlas.brainrender.util import get_color
from neuralib.atlas.typing import Source
from neuralib.atlas.util import roi_points_converter
from neuralib.util.color_logging import setup_clogger

__all__ = [
    'ROI_COLORS',
    'RoiType',
    'BrainReconstructor',
]

ROI_COLORS = ['magenta', 'gold', 'grey']
DEFAULT_REGION_COLORS = ['lightblue', 'pink', 'turquoise']
CAMERA_ANGLE_TYPE = Literal['sagittal', 'sagittal2', 'frontal', 'top', 'top_side', 'three_quarters']
SHADER_STYLE_TYPE = Literal['metallic', 'cartoon', 'plastic', 'shiny', 'glossy']

RoiType: TypeAlias = list[Union[list[np.ndarray], np.ndarray]]
"""RoiType for rendering points"""

Logger = setup_clogger(caller_name=Path(__file__).name)


class CameraAngle(TypedDict):
    """if customized"""
    pos: tuple[int, int, int]
    viewup: tuple[int, int, int]
    clippingRange: tuple[int, int]


class BrainReconstructor(AbstractParser):
    """Wrapper for brainrender 3D reconstruction"""
    DESCRIPTION = 'reconstruct a 3D brain view used brainrender module'
    EPILOG = """
        Data shape of points.
        (N, 3) floating matrix with row (x, y, z), or (N, 4) with row (x, y, z, r),
        where r indicate the index (start from 1, 0 means non-region) in --regions.
        Correspond points and region will use the same color.
        """

    title: str = argument('-T', '--title', metavar='TEXT', default=None)
    source: str = argument('--source', default='allen_mouse_10um', help='atlas source name. allen_human_500um as human')

    # scene
    no_root: bool = argument('--no-root', help='render without root(brain) mesh')

    # source notation
    with_source: bool = argument('-S', '--src', help='whether draw the location for source (experiment dependent)')

    # points
    csv_file: Optional[Path] = argument('-F', '--file', type=Path, default=None, help='csv file')
    points_file: list[Path] = argument('--points', metavar='FILE', type=Path, default=[], action='append')
    radius: float = argument('--roi-radius', default=30, help='each roi radius')
    output: Path = argument('-O', '--out', default=None, help='output path for the html, if None, preview')

    # region
    regions: str = argument('-R', '--region', metavar='NAME,...', type=str_tuple_type, default=[])
    region_colors: str = argument('-C', '--color', metavar='COLOR,...', type=str_tuple_type, default=None)
    regions_alpha: float = argument('--region-alpha', type=float, default=0.35, help='region alpha')
    hemisphere: Literal['right', 'left', 'both'] = argument('-H', '--hemisphere', default='both',
                                                            help='which hemisphere')

    #
    video_output: Optional[Path] = argument('-V', '--video-output', default=None, help='video output')

    # settings
    background: Literal['white', 'black'] = argument('--bg', default='white', help='background color')
    camera_angle: CAMERA_ANGLE_TYPE = argument('--camera', default='three_quarters')
    shader_style: SHADER_STYLE_TYPE = argument('--style', default='plastic')

    #
    scene: brainrender.Scene

    def __init__(self):
        self._need_close_file: list[NamedTemporaryFile] = []
        self.points_list: list[str] = []

    def _render_settings(self):
        from brainrender import settings
        settings.BACKGROUND_COLOR = self.background
        settings.DEFAULT_ATLAS = self.source
        settings.ROOT_ALPHA = 0.35 if self.background == 'black' else 0.2
        settings.SHOW_AXES = False
        settings.WHOLE_SCREEN = False
        settings.DEFAULT_CAMERA = self.camera_angle
        settings.SHADER_STYLE = self.shader_style

        settings.vsettings.screenshot_transparent_background = True
        settings.vsettings.use_fxaa = False

    def post_parsing(self):
        if self.video_output is not None:
            self.source = 'allen_mouse_25um'  # force set for whole brain scene

        self._render_settings()

    def run(self):
        self.scene = brainrender.Scene(root=not self.no_root, inset=False, title=self.title, screenshots_folder='.')
        self.load()
        self.add_points_from_file()
        self.reconstruct()

        #
        if self.video_output is not None:
            self.video_maker(self.video_output)
        elif self.output is not None:
            self.export(self)
        else:
            print('render...')
            self.scene.render()

        #
        if len(self._need_close_file) != 0:
            for f in self._need_close_file:
                f.close()
                Path(f.name).unlink(missing_ok=True)  # winOS

    @property
    def n_files(self) -> int:
        return len(self.points_file)

    # ====== #
    # Points #
    # ====== #

    def add_points(self, rois_list: RoiType):
        for p in rois_list:
            if isinstance(p, np.ndarray):
                # create temporal file in memory for p
                # os handle for NamedTemporaryFile, https://stackoverflow.com/a/23212515
                if sys.platform == 'win32':
                    delete = False
                else:
                    delete = True
                f = NamedTemporaryFile(prefix='.temp-run-3d-proj-', suffix='.npy', delete=delete)

                np.save(f, p)
                f.seek(0)
                self.points_list.append(f.name)
                self._need_close_file.append(f)

    def add_points_from_file(self):
        if self.n_files != 0:
            for it in self.points_file:
                if it.suffix not in ('npy', 'npz'):
                    raise ValueError(f'invalid suffix: {it.suffix}')
                self.points_list.append(str(it))

    def load(self):
        """Overwrite by children"""
        pass

    # =========== #
    # Reconstruct #
    # =========== #

    def reconstruct(self):
        if self.with_source:
            self._reconstruct_source()

        self._reconstruct_region()
        self._reconstruct_points_from_file()

    def _reconstruct_source(self):
        """Depending on the experimental purpose, i.e., viral injection site, targeted location..."""
        # TODO generalize
        src: dict[Source, np.ndarray] = {
            'aRSC': np.array([-1.5, 1, 0.4]),
            'pRSC': np.array([-3.2, 0.8, 0.4])
        }

        color: dict[Source, str] = {
            'aRSC': 'gold',
            'pRSC': 'violet'
        }

        for source, coords in src.items():
            points = roi_points_converter(coords)
            self.scene.add(Points(points, colors=color[source], radius=120))

    def _reconstruct_region(self):
        if self.region_colors is not None:
            color_list = self.region_colors
        else:
            color_list = DEFAULT_REGION_COLORS

        if len(self.regions) != 0:
            for i, region in enumerate(self.regions):
                try:
                    color = color_list[i]
                except IndexError:
                    color = get_color(i, [''])

                Logger.info(f'Plot Rois File: {i}, {region}, {color}')
                self.scene.add_brain_region(region, color=color, alpha=self.regions_alpha, hemisphere=self.hemisphere)

    def _reconstruct_points_from_file(self):
        for i, file in enumerate(self.points_list):
            data = np.load(file)

            if data.ndim != 2:
                raise ValueError(f'wrong dimension: {data.shape}')
            #
            if data.shape[1] == 3:
                colors = get_color(i, ROI_COLORS)
                Logger.info(f'Plot Rois File: {i}, {file}, {colors}')
                self.scene.add(Points(data, name='roi', colors=colors, alpha=0.9, radius=self.radius))

            elif data.shape[1] == 4:  # TODO not test yet
                k = data[:, 3].astype(int)
                for t in np.unique(k):
                    self.scene.add(Points(
                        data[k == t, 0:3],
                        name='rois',
                        colors=get_color(t, ROI_COLORS),
                        alpha=0.6,
                        radius=self.radius
                    ))
            else:
                raise ValueError(f'wrong shape: {data.shape}: {file}')

    @classmethod
    def export(cls, reconstructor: Optional['BrainReconstructor'],
               output: Path = None,
               areas: list[str] = None,
               alpha: float = 0.15):
        """
        export reconstruction as html
        TODO check export / view were different hemispheres, seems opposite in export..

        :param reconstructor: `BrainRenderReconstructor` if use the current scene, and --output cli.
                    Otherwise, general func usage
        :param output:
        :param areas: list of area(s)
        :param alpha:
        :return:
        """

        if isinstance(reconstructor, BrainReconstructor):
            scene = reconstructor.scene
            output = reconstructor.output
        else:
            scene = brainrender.Scene(inset=False, title='', screenshots_folder='.')
            output = output

        if areas is not None:
            if not isinstance(areas, list):
                raise TypeError('')

            for it in areas:
                scene.add_brain_region(it, alpha=alpha)

        scene.export(output)

    def video_maker(self, output_file: Path):
        from brainrender import VideoMaker
        import vedo
        #
        vedo.settings.default_backend = 'vtk'

        #
        d, f = output_file.parent, output_file.stem
        vm = VideoMaker(self.scene, save_fld=d, name=f)
        vm.make_video(azimuth=1, elevation=0, roll=0)

    def get_atlas_brain_globe(self, check_latest=False) -> BrainGlobeAtlas:
        return BrainGlobeAtlas(
            self.source,
            check_latest=check_latest
        )


if __name__ == '__main__':
    BrainReconstructor().main()



"""
RasterMap
=================

Rastermap is a discovery algorithm for neural data. The algorithm was written by Carsen Stringer and Marius Pachitariu

.. seealso:: `<https://github.com/MouseLand/rastermap>`_
.. seealso:: `<https://colab.research.google.com/github/MouseLand/rastermap/blob/main/notebooks/rastermap_largescale.ipynb>`_
.. seealso:: `<https://colab.research.google.com/github/MouseLand/rastermap/blob/main/notebooks/rastermap_widefield.ipynb>`_

"""

from .core import *
from .wfield import *



import abc
from pathlib import Path
from typing import TypedDict, Literal, Union, Optional, final, Any

import attrs
import cellpose.gui.gui
import cv2
import napari
import numpy as np
from typing_extensions import Self

from neuralib.argp import argument, as_argument
from neuralib.segmentation.base import AbstractSegmentationOption
from neuralib.typing import PathLike
from neuralib.util.verbose import fprint

__all__ = ['CPOSE_MODEL',
           'AbstractSegmentationOption',
           'AbstractCellPoseOption',
           'CellPoseEvalResult']


class ChannelDict(TypedDict, total=False):
    none: int
    gray: int
    red: int
    green: int
    blue: int


CELLPOSE_CHANNEL_DICT: ChannelDict = {
    'none': -1,
    'gray': 0,
    'red': 1,
    'green': 2,
    'blue': 3
}

CPOSE_MODEL = Literal['cyto', 'cyto2', 'cyto3']


class AbstractCellPoseOption(AbstractSegmentationOption, metaclass=abc.ABCMeta):
    DESCRIPTION = 'ABC for GUI Cellpose'

    model: CPOSE_MODEL = as_argument(AbstractSegmentationOption.model).with_options(default='cyto3')

    chan_seg: int = argument(
        '-C', '--chan',
        default=CELLPOSE_CHANNEL_DICT['gray'],
        help=f'channel for segmentation default:{CELLPOSE_CHANNEL_DICT}'
    )

    chan_nuclear: int = argument(
        '-N', '-nuclear',
        default=CELLPOSE_CHANNEL_DICT['gray'],
        help='nuclear channel'
    )

    diameter: int = argument(
        '--diameter',
        default=7,
        help='diameter for each neuron (number of each pixel)'
    )

    cellpose_view: bool = argument(
        '--cp', '--cpose',
        help='launch cellpose gui for the analyzed result'
    )

    def seg_output(self, filepath: Path) -> Path:
        return filepath.with_name(filepath.stem + '_seg').with_suffix('.npy')

    # noinspection PyTypeChecker
    def launch_napari(self):
        file = self.seg_output(self.file)
        if not file.exists() or self.force_re_eval:
            self.eval()

        res = CellPoseEvalResult.load(file)

        viewer = napari.Viewer()
        viewer.add_image(res.image, name='image', colormap='cyan')
        viewer.add_image(res.nan_masks(), name='mask', colormap='twilight_shifted', opacity=0.5)
        viewer.add_image(res.nan_outlines(), name='outline', opacity=0.5)

        napari.run()

    def launch_cellpose_gui(self):
        """AttributeError: 'MainW' object has no attribute 'load_3D'. Cellpose version 3.0.10.

        TODO open issue in cellpose -> move ``load_3D`` instance attribute to line above ``io._load_image()`` in ``MainW.__init__()``
        """

        file = self.seg_output(self.file)
        if not file.exists() or self.force_re_eval:
            self.eval()

        cellpose.gui.gui.run(image=str(self.file))  # finding seg result in the same dir


# ========== #
# EvalResult #
# ========== #
# TODO check doc and type in up-to date version
# TODO 3D image need to be tested

class NormParams(TypedDict):
    lowhigh: Optional[Any]
    percentile: list[float, float]
    normalize: bool
    norm3D: bool
    sharpen_radius: float
    smooth_radius: float
    tile_norm_blocksize: float
    tile_norm_smooth3D: float
    invert: bool


@final
@attrs.define
class CellPoseEvalResult:
    """
    Cellpose results

    `Dimension parameters`:

        N = Number of segmented cell

        W = Image width

        H = Image height
    """

    # ====== #
    # Inputs #
    # ====== #

    filename: str
    """image file name"""

    image: Union[np.ndarray, list[np.ndarray]]
    """image array"""

    diameter: float
    """neuronal diameter"""

    chan_choose: list[int]
    """[chan_seg, chan_nuclear]"""

    # ======= #
    # Results #
    # ======= #

    masks: np.ndarray
    """each pixel in the image is assigned to an ROI (H, W)
    list of 2D arrays, labelled image, where 0=no masks; 1,2,...=mask labels
    """

    flows: list[np.ndarray] = attrs.Factory(list)
    """
    flows[0] is XY flow in RGB, 
    flows[1] is the cell probability in range 0-255 instead of 0.0 to 1.0, 
    flows[2] is Z flow in range 0-255 (if it exists, otherwise zeros),
    flows[3] is [dY, dX, cellprob] (or [dZ, dY, dX, cellprob] for 3D),
    flows[4] is pixel destinations (for internal use)
    """

    styles: list[np.ndarray] = attrs.Factory(list)
    """list of 1D arrays of length 256, 
    style vector summarizing each image, also used to estimate size of objects in image"""

    # ==================== #
    # Optional & Overwrite #
    # ==================== #

    # GUI dependent
    colors: Optional[np.ndarray] = attrs.field(default=None, kw_only=True)
    """colors for ROIs (N, 3)"""

    manual_changes: Optional[list[Any]] = attrs.field(default=None, kw_only=True)

    # CLI dependent
    est_diam: Optional[float] = attrs.field(default=None, kw_only=True)
    """estimated diameter (if run on command line)"""

    #
    model_path: int = attrs.field(default=0, kw_only=True)
    flow_threshold: Optional[float] = attrs.field(default=None, kw_only=True)
    cellprob_threshold: float = attrs.field(default=0, kw_only=True)
    normalize_params: Optional[NormParams] = attrs.field(default=None, kw_only=True)

    # restore
    img_restore: Optional[list[np.ndarray]] = attrs.field(default=None, kw_only=True)
    restore: Optional[str] = attrs.field(default=None, kw_only=True)
    ratio: float = attrs.field(default=1.0, kw_only=True)

    # Overwrite while calling save
    outlines: np.ndarray = attrs.field(default=np.array([]), kw_only=True)
    """outlines of ROIs. `Array[uint16, [H, W]]`"""

    ismanual: np.ndarray = attrs.field(default=np.array([]), kw_only=True)
    """whether or not mask k was manually drawn or computed by the cellpose algorithm. `Array[bool, N]`"""

    @classmethod
    def load(cls, seg_file: PathLike) -> Self:
        if not isinstance(seg_file, Path):
            seg_file = Path(seg_file)

        res = np.load(seg_file, allow_pickle=True).item()

        image_file = Path(res['filename'])
        if not image_file.exists():
            fprint(f'No image data found {image_file}', vtype='warning')
            image = None
        else:
            image = cv2.imread(res['filename'])

        return cls(**res, image=image)

    def save_seg_file(self, image_file: str) -> None:
        """Save as ``seg.npy`` file`"""
        from cellpose.io import masks_flows_to_seg

        if not isinstance(self.image, list):
            self.image = [self.image]

        masks_flows_to_seg(self.image,
                           self.masks,
                           self.flows,
                           image_file,
                           diams=self.diameter,
                           imgs_restore=self.img_restore,
                           restore_type=self.restore,
                           ratio=self.ratio)

    def save_roi(self, output_file: PathLike) -> None:
        """Save as imageJ ``.roi`` file.
        CHECKOUT native BUG: `<https://github.com/MouseLand/cellpose/issues/969>`_
        """
        from roifile import ImagejRoi, ROI_TYPE, ROI_OPTIONS

        points = np.fliplr(self.points)  # XY rotate in .roi format
        roi = ImagejRoi(
            roitype=ROI_TYPE.POINT,
            options=ROI_OPTIONS.PROMPT_BEFORE_DELETING | ROI_OPTIONS.SUB_PIXEL_RESOLUTION,
            n_coordinates=self.points.shape[0],
            integer_coordinates=points,
            subpixel_coordinates=points
        )

        roi.tofile(output_file)

    def nan_masks(self) -> np.ndarray:
        """value 0 in ``masks`` to nan"""
        masks = self.masks.copy().astype(np.float_)
        masks[masks == 0] = np.nan

        return masks

    def nan_outlines(self) -> np.ndarray:
        """value 0 in ``outlines`` to nan"""
        outlines = self.outlines.copy().astype(np.float_)
        outlines[outlines == 0] = np.nan

        return outlines

    @property
    def points(self) -> np.ndarray:
        """Calculate center of each segmented area in pixel. `Array[int, N]`"""
        labels = np.unique(self.masks)
        labels = labels[labels != 0]  # remove background

        n_neurons = len(labels)
        centers = np.zeros((n_neurons, 2))
        for i, label in enumerate(labels):
            segment_coords = np.argwhere(self.masks == label)
            center = segment_coords.mean(axis=0)
            centers[i] = center

        return np.round(centers).astype(int)



"""
Suite2p Result Parser & Utils
===============================

:author:
    Yu-Ting Wei

This module provide the container for suite2p registration/segmentation 2P results.

Also include utilities for signal processing and basic plotting

"""
from .core import *
from .plot import *
from .signals import *



"""
StarDist
=========


**Example of run an image segmentation in 2D mode and visualize using napari**

.. code-block:: console

    python -m neuralib.segmentation.stardist.run_2d -F <IMAGE_FILE> --napari

- It produces an output ``<IMAGE_FILE>.npz`` in the same directory, If the output exists, the napari will directly use the output. If need run again and rewrite the output, use ``--force-eval`` option



**Example of run batch image segmentation in 2D mode from a directory (.tif files)**

.. code-block:: console

    python -m neuralib.segmentation.stardist.run_2d -D <DIRECTORY> --suffix .tif

- It produces multiple ``*.npz`` in the same directory.


**See help using** ``-h`` **option**

.. code-block:: console

    python -m neuralib.segmentation.stardist.run_2d -h

"""



from __future__ import annotations

import shutil
from pathlib import Path
from typing import TypedDict, Final, Literal, get_args, Iterable

import attrs
import numpy as np
import polars as pl
from typing_extensions import Self, TypeAlias

from neuralib.atlas.ccf.core import (
    AbstractCCFDir,
    SagittalCCFDir,
    SagittalCCFOverlapDir,
    CoronalCCFOverlapDir,
    CoronalCCFDir
)
from neuralib.atlas.ccf.norm import MouseBrainRoiNormHandler, ROIS_NORM_TYPE
from neuralib.atlas.map import merge_until_level, NUM_MERGE_LAYER, DEFAULT_FAMILY_DICT
from neuralib.atlas.typing import Area, HEMISPHERE_TYPE, Source, Channel
from neuralib.atlas.util import PLANE_TYPE
from neuralib.util.color_logging import setup_clogger, LOGGING_IO_LEVEL
from neuralib.util.utils import uglob

__all__ = [
    'FluorReprType',
    #
    'UserInjectionConfig',
    'CLASSIFIED_COL',
    'RoiClassifier',
    'RoiClassifiedNormTable',
    #
    'supply_overlap_dataframe',
    #
    '_concat_channel',
    'parse_csv'
]

Logger = setup_clogger(caller_name=Path(__file__).name)

FluorReprType: TypeAlias = dict[Channel, Source]


# ======= #
# Configs #
# ======= #

class UserInjectionConfig(TypedDict):
    area: Area
    """injection area"""
    hemisphere: HEMISPHERE_TYPE
    """injection hemisphere"""
    ignore: bool
    """whether local roi counts will be ignored"""
    fluor_repr: FluorReprType
    """fluorescence color and tracing source alias pairs"""


# Example (replace to user-specific)
_DEFAULT_RSP_CONFIG = UserInjectionConfig(
    area='RSP',
    hemisphere='ipsi',
    ignore=True,
    fluor_repr=dict(
        rfp='pRSC',
        gfp='aRSC',
        overlap='overlap'
    )
)

# ========== #
# Classifier #
# ========== #

CLASSIFIED_COL = Literal[
    'acronym_abbr',
    'merge_ac_0',
    'merge_ac_1',
    'merge_ac_2',
    'merge_ac_3',
    'merge_ac_4',
    'family'
]


class RoiClassifier:
    __slots__ = ('ccf_dir', 'merge_level', 'plane',
                 'ignore_injection_site', 'fluor_repr', '_fluor_order',
                 '_injection_area', '_injection_hemi', '_parse_df')

    def __init__(self, ccf_dir: AbstractCCFDir,
                 merge_level: int | str | None = None,
                 plane: PLANE_TYPE = 'coronal',
                 config: UserInjectionConfig | None = None):
        """
        Classifier for see the ROIs distribution across the mouse brain

        :param ccf_dir: ccf directory info
        :param merge_level: which number (int) of hierarchical merge area, or the col name (str) for output generate
        :param plane: slice cutting orientation
        :param config: `UserInjectionConfig` for customized injection info
        """

        self.ccf_dir = ccf_dir
        self.merge_level = merge_level
        self.plane = plane

        # config
        if config is None:
            config = _DEFAULT_RSP_CONFIG

        self.ignore_injection_site: Final[bool] = config['ignore']
        self.fluor_repr: FluorReprType = config['fluor_repr']  # constructs can be swapped if experimental need

        self._fluor_order: Final[tuple[Channel, ...]] = tuple(list(config['fluor_repr'].keys()))
        self._injection_area: Final[Area] = config['area']  # use startswith to ignore
        self._injection_hemi: Final[HEMISPHERE_TYPE] = config['hemisphere']

        # cache
        self._parse_df: pl.DataFrame | None = None

    @property
    def has_overlap(self) -> bool:
        """If has overlap channel counts"""
        config_check = 'overlap' in self.fluor_repr
        if self.ccf_dir.with_overlap_sources != config_check:
            raise RuntimeError('check UserInjectionConfig and AbstractCCFDir')
        return config_check

    @property
    def parsed_df(self) -> pl.DataFrame:
        """dataframe after parsing

        Example::

            ┌───────────────────────────────────┬─────────┬─────────────┬─────────────┬─────────────┬─────────┬─────────┬────────┬───────────────────────────┬──────────────┬────────┬────────────┬────────────┬────────────┬────────────┬────────────┬───────────┐
            │ name                              ┆ acronym ┆ AP_location ┆ DV_location ┆ ML_location ┆ avIndex ┆ channel ┆ source ┆ abbr                      ┆ acronym_abbr ┆ hemi.  ┆ merge_ac_0 ┆ merge_ac_1 ┆ merge_ac_2 ┆ merge_ac_3 ┆ merge_ac_4 ┆ family    │
            │ ---                               ┆ ---     ┆ ---         ┆ ---         ┆ ---         ┆ ---     ┆ ---     ┆ ---    ┆ ---                       ┆ ---          ┆ ---    ┆ ---        ┆ ---        ┆ ---        ┆ ---        ┆ ---        ┆ ---       │
            │ str                               ┆ str     ┆ f64         ┆ f64         ┆ f64         ┆ i64     ┆ str     ┆ str    ┆ str                       ┆ str          ┆ str    ┆ str        ┆ str        ┆ str        ┆ str        ┆ str        ┆ str       │
            ╞═══════════════════════════════════╪═════════╪═════════════╪═════════════╪═════════════╪═════════╪═════════╪════════╪═══════════════════════════╪══════════════╪════════╪════════════╪════════════╪════════════╪════════════╪════════════╪═══════════╡
            │ Ectorhinal area/Layer 5           ┆ ECT5    ┆ -3.03       ┆ 4.34        ┆ -4.5        ┆ 377     ┆ gfp     ┆ aRSC   ┆ Ectorhinal area           ┆ ECT          ┆ contra ┆ ECT        ┆ ECT        ┆ ECT        ┆ ECT        ┆ ECT        ┆ ISOCORTEX │
            │ Perirhinal area layer 6a          ┆ PERI6a  ┆ -3.03       ┆ 4.42        ┆ -4.37       ┆ 372     ┆ gfp     ┆ aRSC   ┆ Perirhinal area           ┆ PERI         ┆ contra ┆ PERI       ┆ PERI       ┆ PERI       ┆ PERI       ┆ PERI       ┆ ISOCORTEX │
            │ …                                 ┆ …       ┆ …           ┆ …           ┆ …           ┆ …       ┆ …       ┆ …      ┆ …                         ┆ …            ┆ …      ┆ …          ┆ …          ┆ …          ┆ …          ┆ …          ┆ …         │
            │ Ventral auditory area layer 6a    ┆ AUDv6a  ┆ -2.91       ┆ 3.52        ┆ 4.46        ┆ 156     ┆ rfp     ┆ pRSC   ┆ Ventral auditory area     ┆ AUDv         ┆ ipsi   ┆ AUD        ┆ AUD        ┆ AUD        ┆ AUD        ┆ AUDv       ┆ ISOCORTEX │
            │ Ectorhinal area/Layer 6a          ┆ ECT6a   ┆ -2.91       ┆ 4.14        ┆ 4.47        ┆ 378     ┆ rfp     ┆ pRSC   ┆ Ectorhinal area           ┆ ECT          ┆ ipsi   ┆ ECT        ┆ ECT        ┆ ECT        ┆ ECT        ┆ ECT        ┆ ISOCORTEX │
            │ Temporal association areas layer… ┆ TEa5    ┆ -2.91       ┆ 4.02        ┆ 4.55        ┆ 365     ┆ rfp     ┆ pRSC   ┆ Temporal association area ┆ TEa          ┆ ipsi   ┆ TEa        ┆ TEa        ┆ TEa        ┆ TEa        ┆ TEa        ┆ ISOCORTEX │
            └───────────────────────────────────┴─────────┴─────────────┴─────────────┴─────────────┴─────────┴─────────┴────────┴───────────────────────────┴──────────────┴────────┴────────────┴────────────┴────────────┴────────────┴────────────┴───────────┘
        """
        if self._parse_df is None:
            df = self._cache_parsed_dataframe()

            if self.ignore_injection_site:
                df = self._ignore_injection_site(df)
            self._parse_df = df

        return self._parse_df

    @parsed_df.setter
    def parsed_df(self, df: pl.DataFrame):
        """For supply overlap purpose"""
        self._parse_df = df

    # ============== #
    # Pre-Processing #
    # ============== #

    def _cache_parsed_dataframe(self, force_save: bool = False) -> pl.DataFrame:
        """concat & parse & add fields and cache a parsed dataframe"""
        file = self.ccf_dir.parse_csv

        if not file.exists() or force_save:
            df = parse_csv(self.ccf_dir, self.fluor_repr, plane=self.plane)

            ac = df['acronym']

            # add merge level cols
            df = df.with_columns(
                pl.Series(name=f'merge_ac_{level}', values=merge_until_level(ac, level))
                for level in range(NUM_MERGE_LAYER)
            )

            # add family col
            def categorize_family(row) -> str:
                for name, family in DEFAULT_FAMILY_DICT.items():
                    if row in family:
                        return name

            df = df.with_columns(pl.col('merge_ac_0')
                                 .map_elements(categorize_family, return_dtype=pl.Utf8)
                                 .fill_null('unknown')
                                 .alias('family'))

            df.write_csv(file)
        else:
            df = pl.read_csv(file)

        return df

    def _ignore_injection_site(self, df: pl.DataFrame, verbose: bool = True) -> pl.DataFrame:
        area = self._injection_area
        hemi = self._injection_hemi
        Logger.info(f'remove {area} in {hemi} in {self.ccf_dir.animal}')

        expr1 = pl.col('acronym').str.starts_with(area)
        expr2 = pl.col('hemi.') == hemi

        if verbose:
            exc = (
                df.filter(pl.col('acronym').str.starts_with(area))
                .group_by('hemi.').agg(pl.len())
            )

            print(exc)

        return df.filter(~(expr1 & expr2))

    # ===== #
    # Utils #
    # ===== #

    # noinspection PyTypeChecker
    @property
    def classified_column(self) -> CLASSIFIED_COL:
        """ """
        if self.merge_level is None:
            return 'acronym_abbr'

        if isinstance(self.merge_level, int):
            return f'merge_ac_{self.merge_level}'
        elif isinstance(self.merge_level, str) and self.merge_level in self.parsed_df.columns:
            return self.merge_level
        else:
            raise ValueError(f'invalid merge level: {self.merge_level}')

    @property
    def channels(self) -> list[Channel]:
        """list of fluorescence channels
        (sorted based on :class:`~neuralib.atlas.ccf.classifier.UserInjectionConfig.fluor_repr`)"""
        chan_list = self.parsed_df['channel'].unique()
        return sorted(chan_list, key=lambda it: self._fluor_order.index(it))

    @property
    def n_channels(self) -> int:
        """number of fluorescence channels"""
        return len(self.channels)

    @property
    def sources(self) -> list[Source]:
        """list of unique sources"""
        return list(self.parsed_df['source'].unique().sort())

    @property
    def n_sources(self) -> int:
        """number of source name"""
        return len(self.sources)

    @property
    def areas(self) -> list[Area]:
        """list of area in the given ``classified_column``"""
        return list(self.parsed_df[self.classified_column].unique())

    @property
    def n_total_rois(self) -> int:
        """number of total rois"""
        return self.parsed_df.shape[0]

    @property
    def nroi_channel(self) -> pl.DataFrame:
        """channels and counts Dataframe"""
        return self.parsed_df['channel'].value_counts()

    @property
    def nroi_source(self) -> pl.DataFrame:
        """channels and counts Dataframe"""
        return self.parsed_df['source'].value_counts()

    @property
    def nroi_area(self) -> pl.DataFrame:
        """dict with key (area name) and value (number of cells)"""
        return self.parsed_df[self.classified_column].value_counts()

    # ========================= #
    # Classifier Output Methods #
    # ========================= #

    def get_percent_sorted_df(self, hemisphere: HEMISPHERE_TYPE = 'both') -> pl.DataFrame:
        """
        Get percentage-sorted dataframe

        :param hemisphere: which hemisphere {'ipsi', 'contra', 'both'}
        :return: dataframe

        **hemi**: optional col. if hemisphere is`both`, then sum together

        Example of merge level equal to 2::

            ┌─────────┬────────────┬────────┬────────┬───────────┐
            │ source  ┆ merge_ac_2 ┆ *hemi. ┆ n_rois ┆ percent   │
            │ ---     ┆ ---        ┆ ---    ┆ ---    ┆ ---       │
            │ str     ┆ str        ┆ str    ┆ u32    ┆ f64       │
            ╞═════════╪════════════╪════════╪════════╪═══════════╡
            │ pRSC    ┆ VIS        ┆ ipsi   ┆ 5701   ┆ 22.761209 │
            │ aRSC    ┆ MO         ┆ ipsi   ┆ 2277   ┆ 21.360225 │
            │ overlap ┆ MO         ┆ ipsi   ┆ 571    ┆ 17.799252 │
            │ aRSC    ┆ ACA        ┆ ipsi   ┆ 1729   ┆ 16.219512 │
            │ overlap ┆ SUB        ┆ ipsi   ┆ 494    ┆ 15.399002 │
            │ …       ┆ …          ┆ …      ┆ …      ┆ …         │
            │ pRSC    ┆ MED        ┆ contra ┆ 1      ┆ 0.003992  │
            │ pRSC    ┆ PPN        ┆ contra ┆ 1      ┆ 0.003992  │
            │ pRSC    ┆ APr        ┆ ipsi   ┆ 1      ┆ 0.003992  │
            │ pRSC    ┆ BMA        ┆ contra ┆ 1      ┆ 0.003992  │
            │ pRSC    ┆ P5         ┆ contra ┆ 1      ┆ 0.003992  │
            └─────────┴────────────┴────────┴────────┴───────────┘
        """

        if hemisphere in ('contra', 'ipsi'):
            cols = ['source', self.classified_column, 'hemi.']
        else:
            cols = ['source', self.classified_column]

        df = self.parsed_df
        df = (df.select(cols)
              .group_by(cols)
              .agg(pl.col(self.classified_column).count().alias('n_rois'))
              .with_columns((pl.col('n_rois') / pl.col('n_rois').sum().over('source') * 100).alias('percent'))
              .sort('percent', descending=True))

        return df

    def get_classified_data(
            self,
            norm: MouseBrainRoiNormHandler | None = None, *,
            top_area: int | None = None,
            source: Source | None = None,
            add_other: bool = False,
            supply_overlap: bool = True,
            hemisphere: HEMISPHERE_TYPE = 'both',
            area: Area | list[Area] | None = None
    ) -> RoiClassifiedNormTable:
        """
        processed data for plotting / visualization

        :param norm: :class:`BrainMapNormHandler`
        :param top_area: select top ranks area based on channel-based normalized percentage
        :param source: specify source, if not then produce all source
        :param add_other: the rest of regions (after top selection), classified as `other` (i.e., pie chart)
        :param supply_overlap: add overlap roi counts into other channel(s)
        :param hemisphere: filter the output data with selected hemisphere {'ipsi', 'contra', 'both'}
        :param area: filter the output data with selected area
        :return: :class:`RoiClassifiedNormTable`
        """
        if supply_overlap and source != 'overlap':
            sup_source = list(self.fluor_repr.values())
            sup_source.remove('overlap')
            supply_df = supply_overlap_dataframe(self.parsed_df, supply_channel_name=sup_source)
            self.parsed_df = supply_df  # trigger setter
        else:
            Logger.warning('Source counts exclude overlap channel!')

        df = self.get_percent_sorted_df(hemisphere)

        #
        if source is not None:
            df = df.filter(pl.col('source') == source)

        #
        if top_area is not None:
            top_area = int(top_area)
            if source is not None:
                df = self._select_top_region_single_channel(df, top_area, source, add_other)
            else:
                df = self._select_top_region_all_channel(df, top_area)

        #
        if norm is not None:
            if norm.norm_type is not None:
                df = norm.expand(df, area_col=self.classified_column, expand_cols=['n_rois'])

            if norm.norm_type in ('cell', 'volume'):
                df = norm.handle_failure(df)

        norm_type = norm.norm_type if norm is not None else None

        #
        df = df.with_columns(pl.lit(self.ccf_dir.animal).alias('animal'))

        #
        ret = RoiClassifiedNormTable(self, norm_type, hemisphere, df)

        if hemisphere != 'both':
            Logger.info(f'Filter classified result from: {hemisphere} hemisphere')
            ret = ret.with_hemisphere(hemisphere)

        if area is not None:
            ret = ret.with_areas(area)

        return ret

    def _select_top_region_single_channel(self,
                                          df: pl.DataFrame,
                                          top_area: int,
                                          source: Source,
                                          add_other: bool = False) -> pl.DataFrame:
        if top_area > df.shape[0]:
            Logger.warning(f'only {df.shape[0]} areas are classified,'
                           f'{top_area} areas exceed, thus use all areas instead')
        else:
            df = df[:top_area]

        if add_other and (100 - df['percent'].sum()) > 0:
            other_perc = max(0, 100 - df['percent'].sum())
            total_num = self.nroi_source.filter(pl.col('source') == source)['count'].item()
            other_num = total_num - df['n_rois'].sum()

            schema = {
                df.columns[i]: df.dtypes[i]
                for i in range(df.shape[1])
            }
            row = pl.DataFrame([[source, 'other', other_num, other_perc]], schema=schema, orient='row')  # `other` row

            df = pl.concat([df, row])

        return df

    def _select_top_region_all_channel(self, df: pl.DataFrame, top_area: int) -> pl.DataFrame:
        ref_df = df[:top_area]
        _region = ref_df[self.classified_column].unique()

        return df.filter(pl.col(self.classified_column).is_in(_region))


# ================ #
# Classified Table #
# ================ #

@attrs.define
class RoiClassifiedNormTable:
    source_classifier: RoiClassifier

    norm_type: ROIS_NORM_TYPE | None = attrs.field(
        validator=attrs.validators.optional(attrs.validators.in_(get_args(ROIS_NORM_TYPE)))
    )

    hemisphere: HEMISPHERE_TYPE = attrs.field(validator=attrs.validators.in_(get_args(HEMISPHERE_TYPE)))

    data: pl.DataFrame
    """processed data after ROIClassifier

    * Optional col: hemi. , only existed if init with either ``hemisphere`` is 'ipsi' or 'contra' 
    
    * <ROIS_NORM_TYPE>_norm depending on ``norm_type``
    
    Example of data with volume normalization method::
    
        ┌─────────┬────────────┬────────┬───────────┬───┬───────────┬───────────┬─────────────────┬────────┐
        │ channel ┆ merge_ac_2 ┆ n_rois ┆ percent   ┆ … ┆ Volumes   ┆ n_neurons ┆ *volume_norm_n_r┆ animal │
        │ ---     ┆ ---        ┆ ---    ┆ ---       ┆   ┆ [mm^3]    ┆ ---       ┆ ois             ┆ ---    │
        │ str     ┆ str        ┆ i64    ┆ f64       ┆   ┆ ---       ┆ i64       ┆ ---             ┆ str    │
        │         ┆            ┆        ┆           ┆   ┆ f64       ┆           ┆ f64             ┆        │
        ╞═════════╪════════════╪════════╪═══════════╪═══╪═══════════╪═══════════╪═════════════════╪════════╡
        │ overlap ┆ ACA        ┆ 423    ┆ 30.344333 ┆ … ┆ 5.222484  ┆ 337372    ┆ 80.995934       ┆ YW051  │
        │ gfp     ┆ MO         ┆ 3352   ┆ 24.545987 ┆ … ┆ 22.248234 ┆ 985411    ┆ 150.663641      ┆ YW051  │
        │ rfp     ┆ ACA        ┆ 1383   ┆ 23.791502 ┆ … ┆ 5.222484  ┆ 337372    ┆ 264.816494      ┆ YW051  │
        │ gfp     ┆ ACA        ┆ 3130   ┆ 22.920328 ┆ … ┆ 5.222484  ┆ 337372    ┆ 599.331616      ┆ YW051  │
        │ …       ┆ …          ┆ …      ┆ …         ┆ … ┆ …         ┆ …         ┆ …               ┆ …      │
        │ overlap ┆ SS         ┆ 1      ┆ 0.071736  ┆ … ┆ 37.177937 ┆ 2384622   ┆ 0.026898        ┆ YW051  │
        │ overlap ┆ ECT        ┆ 1      ┆ 0.071736  ┆ … ┆ 3.457703  ┆ 387378    ┆ 0.289209        ┆ YW051  │
        │ overlap ┆ TEa        ┆ 1      ┆ 0.071736  ┆ … ┆ 3.860953  ┆ 386396    ┆ 0.259003        ┆ YW051  │
        │ rfp     ┆ TT         ┆ 1      ┆ 0.017203  ┆ … ┆ 1.734078  ┆ 124596    ┆ 0.576675        ┆ YW051  │
        └─────────┴────────────┴────────┴───────────┴───┴───────────┴───────────┴─────────────────┴────────┘
    """

    @property
    def value_col(self) -> str:
        """column name for the value"""
        if self.norm_type in ('cell', 'volume'):
            return f'{self.norm_type}_norm_n_rois'
        elif self.norm_type == 'channel':
            return 'percent'
        else:
            return 'n_rois'

    @property
    def classified_column(self) -> CLASSIFIED_COL:
        """classified column based on the ``RoiClassifier``"""
        return self.source_classifier.classified_column

    def with_areas(self, areas: Area | list[Area]) -> Self:
        """
        filter the data with selected area

        :param areas: area or list of area
        :return: :class:`RoiClassifiedNormTable`
        """
        if isinstance(areas, str):
            areas = [areas]

        _data = self.data.filter(pl.col(self.classified_column).is_in(areas))

        if _data.is_empty():
            raise ValueError(f'{areas} not found')

        return attrs.evolve(self, data=_data)

    def with_hemisphere(self, hemi: HEMISPHERE_TYPE) -> Self:
        """
        filter the data with selected hemisphere

        :param hemi: {'ipsi', 'contra'}
        :return: :class:`RoiClassifiedNormTable`
        """
        if hemi not in ('ipsi', 'contra'):
            raise ValueError('')

        return attrs.evolve(self,
                            hemisphere=hemi,
                            data=self.data.filter(pl.col('hemi.') == hemi))

    def get_bias_value_dataframe(self,
                                 yunit: str = 'percent',
                                 to_index=True,
                                 verbose=True) -> pl.DataFrame:
        """
        Dataframe with `bias value` using either

        1. simple subtraction using either *percent* or *n_rois* in `self.data` (bias_value)

        2. index calculation (bias index)

        :param yunit:
        :param to_index: to bias index (1og2 (P_areaA / P_areaB)). refer to Chen et al., 2022. biorxiv.
            otherwise, do simple subtraction
            NOTE that index is currently calculate based on `channel-normalization` value (percent col)
        :param verbose: do print for output
        :return: DataFrame

        Example::

            ┌────────────┬────────────┐
            │ merge_ac_1 ┆ bias_value*│
            │ ---        ┆ ---        │
            │ str        ┆ f64        │
            ╞════════════╪════════════╡
            │ MO         ┆ -18.591053 │
            │ SS         ┆ -5.362179  │
            │ …          ┆ …          │
            │ CP         ┆ 3.897736   │
            │ VIS        ┆ 12.899966  │
            └────────────┴────────────┘
        """
        bias_col = 'bias_value'

        if to_index:
            expr_calc = (pl.col('pRSC') / pl.col('aRSC')).map_elements(np.log2, return_dtype=pl.Float64)
            yunit = 'percent'  # force
            bias_col = 'bias_index'
        else:
            expr_calc = (pl.col('pRSC') - pl.col('aRSC'))

        df = (
            self.data
            .select(self.classified_column, 'source', yunit)
            .sort(self.classified_column, 'source')
            .pivot(values=yunit, index=self.classified_column, on='source', aggregate_function='first')
            .fill_null(0)
            .with_columns(expr_calc.alias(bias_col))
            .filter(~pl.col(bias_col).is_infinite())  # log2 index calc fail
            .select(self.classified_column, bias_col)
            .sort(by=bias_col)
        )

        if verbose:
            from neuralib.util.verbose import printdf
            printdf(df)

        return df

    def to_bokeh(self, yunit: str = 'percent') -> dict[str, list]:
        """
        to bokeh data ColumnSource input structure

        :param yunit: {'percent', 'n_rois'}
        :return:
            key: `x` & `counts`

            value: (area, channel) & neuronal counts or percentage
        """
        if yunit == 'n_rois':
            icol = 2
        elif yunit == 'percent':
            icol = 3
        else:
            raise ValueError(f'unknown yunit: {yunit}')

        data = self.data.sort('source')

        ret = {'x': [], 'counts': []}
        for row in data.rows():
            ret['x'].append((row[1], row[0]))
            ret['counts'].append(row[icol])

        return ret

    def to_winner_dataframe(self) -> pl.DataFrame:
        """
        For ternary plot in plotly module

        :return: dataframe

        ::

            ┌────────────┬─────────┬──────┬──────┬───────┬────────┐
            │ merge_ac_1 ┆ overlap ┆ pRSC ┆ aRSC ┆ total ┆ winner │
            │ ---        ┆ ---     ┆ ---  ┆ ---  ┆ ---   ┆ ---    │
            │ str        ┆ i64     ┆ i64  ┆ i64  ┆ i64   ┆ str    │
            ╞════════════╪═════════╪══════╪══════╪═══════╪════════╡
            │ ACA        ┆ 1000    ┆ 1048 ┆ 2865 ┆ 4913  ┆ gfp    │
            │ RSP        ┆ 327     ┆ 1086 ┆ 1345 ┆ 2758  ┆ gfp    │
            │ …          ┆ …       ┆ …    ┆ …    ┆ …     ┆ …      │
            │ CB         ┆ 0       ┆ 0    ┆ 1    ┆ 1     ┆ gfp    │
            │ CUN        ┆ 0       ┆ 0    ┆ 1    ┆ 1     ┆ gfp    │
            └────────────┴─────────┴──────┴──────┴───────┴────────┘
        """
        df = (self.data
              .pivot(values='n_rois', on='source', index=self.classified_column, aggregate_function='first')
              .fill_nan(0)
              .fill_null(0)
              .with_columns((pl.col('pRSC') + pl.col('aRSC') + pl.col('overlap')).alias('total')))

        cols = self.source_classifier.sources
        region_counts = df.select(cols).to_numpy()
        winner_idx = np.argmax(region_counts, axis=1).astype(int)
        df = df.with_columns(pl.Series([cols[idx] for idx in winner_idx]).alias('winner'))

        return df

    def write_csv(self, out: Path,
                  animal_id_col: str | None = None):
        data = self.data
        if animal_id_col is not None:
            data = data.with_columns(pl.lit(animal_id_col).alias('animal_id'))

        data.write_csv(out)


# ======================================== #
# Concat CSV (Add Source and Channel Info) #
# ======================================== #


# noinspection PyTypeChecker
def _concat_channel(ccf_dir: AbstractCCFDir,
                    fluor_repr: FluorReprType,
                    plane: PLANE_TYPE) -> pl.DataFrame:
    """
    Find the csv data from `labelled_roi_folder`, if multiple files are found, concat to single df.
    `channel` & `source` columns are added to the dataframe.
    
    If sagittal slice, auto move ipsi/contra hemispheres dataset (`resize_ipsi`, `resize_contra`) 
    to new `resize` directory

    :param ccf_dir: :class:`~neuralib.atlas.ccf.core.AbstractCCFDir()`
    :param fluor_repr: ``FluorReprType``
    :param plane: ``PLANE_TYPE`` {'coronal', 'sagittal', 'transverse'}
    :return:
    """
    if plane == 'sagittal':
        _auto_sagittal_combine(ccf_dir)
    elif plane == 'coronal':
        _auto_coronal_combine(ccf_dir)

    f = list(ccf_dir.labelled_roi_folder.glob('*.csv'))

    if len(f) == 1:
        return _single_proc(f, fluor_repr)
    else:
        return _multiple_concat_proc(f, plane, ccf_dir, fluor_repr)


def _auto_overlap_copy(ccf: CoronalCCFOverlapDir | SagittalCCFOverlapDir) -> None:
    src = uglob(ccf.labelled_roi_folder_overlap, '*.csv')
    filename = f'{ccf.animal}_overlap_roitable'
    if ccf.plane_type == 'sagittal':
        filename += f'_{ccf.hemisphere}'

    dst = (ccf.labelled_roi_folder / filename).with_suffix('.csv')
    shutil.copy(src, dst)
    Logger.log(LOGGING_IO_LEVEL, f'copy overlap file from {src} to {dst}')


def _auto_coronal_combine(ccf_dir: CoronalCCFDir | CoronalCCFOverlapDir):
    _auto_overlap_copy(ccf_dir)


def _auto_sagittal_combine(ccf_dir: SagittalCCFDir | SagittalCCFOverlapDir) -> None:
    """copy file from overlap dir to major fluorescence (channel) folder,
    then combine different hemisphere data"""

    old_args = ccf_dir.hemisphere

    def with_hemisphere_stem(ccf: SagittalCCFDir | SagittalCCFOverlapDir) -> list[Path]:
        ls = list(ccf.labelled_roi_folder.glob('*.csv'))
        for it in ls:
            if ccf.hemisphere not in it.name:
                new_path = it.with_stem(it.stem + f'_{ccf.hemisphere}')
                it.rename(new_path)

        return list(ccf.labelled_roi_folder.glob('*.csv'))  # new glob

    mv_list = []

    ccf_dir.hemisphere = 'ipsi'
    if isinstance(ccf_dir, SagittalCCFOverlapDir):
        _auto_overlap_copy(ccf_dir)
    ext = with_hemisphere_stem(ccf_dir)
    mv_list.extend(ext)

    #
    ccf_dir.hemisphere = 'contra'
    if isinstance(ccf_dir, SagittalCCFOverlapDir):
        _auto_overlap_copy(ccf_dir)
    ext = with_hemisphere_stem(ccf_dir)
    mv_list.extend(ext)

    #
    ccf_dir.hemisphere = 'both'  # as resize
    target = ccf_dir.labelled_roi_folder
    for file in mv_list:
        shutil.copy(file, target / file.name)
        Logger.log(LOGGING_IO_LEVEL, f'copy file from {file} to {target / file.name}')

    ccf_dir.hemisphere = old_args  # assign back


def _multiple_concat_proc(f: list[Path],
                          plane: PLANE_TYPE,
                          ccf_dir: AbstractCCFDir,
                          fluor_repr: FluorReprType,
                          strict: bool = True,
                          n_channels: int = 2,
                          with_overlap_counts: bool = True) -> pl.DataFrame:
    """

    :param f:
    :param plane: ``PLANE_TYPE``
    :param ccf_dir:  ``AbstractCCFDir``
    :param fluor_repr:  ``FluorReprType``
    :param strict: if strict check the file number is correct for the pipeline
    :param n_channels: number of fluorescence channel (without overlap)
    :param with_overlap_counts: if has overlap channel counts
    :return:
    """
    Logger.log(LOGGING_IO_LEVEL, f'load multiple csv files: {len(f)} files')

    df_list = []
    for ff in f:
        if 'rfp' in ff.name:
            channel = 'rfp'
        elif 'gfp' in ff.name:
            channel = 'gfp'
        elif 'overlap' in ff.name:
            channel = 'overlap'
        else:
            continue

        source = fluor_repr[channel]

        df_list.append(pl.read_csv(ff).with_columns(pl.lit(channel).alias('channel'))
                       .with_columns(pl.lit(source).alias('source')))

    # validate numbers of file
    if strict:
        exp = n_channels

        if with_overlap_counts:
            exp += 1

        if plane == 'sagittal':
            exp *= 2

        if len(df_list) != exp:
            raise RuntimeError(f'missing csv in {ccf_dir.labelled_roi_folder} for {plane} pipeline')

    return pl.concat(df_list)


def _single_proc(f: list[Path], fluor_repr: FluorReprType):
    Logger.log(LOGGING_IO_LEVEL, 'load single csv file')
    file = f[0]
    df = pl.read_csv(file)
    for pattern in ['rfp', 'gfp', 'overlap']:
        if pattern in file.name:
            channel = pattern
            source = fluor_repr[channel]

            return (df.with_columns(pl.lit(channel).alias('channel'))
                    .with_columns(pl.lit(source).alias('source')))

    raise RuntimeError('')


# ========================= #
# Raw CSV Parsing Functions #
# ========================= #

def parse_csv(ccf_dir: AbstractCCFDir | None,
              fluor_repr: FluorReprType, *,
              plane: PLANE_TYPE = 'coronal',
              df: pl.DataFrame | None = None,
              invert_hemi: bool = False) -> pl.DataFrame:
    """
    Narrow down the info in the ccf roi output

    columns `abbr`, `acronym_abbr` and `hemi` are added to the df

    :param ccf_dir: if None, directly give ``df`` arg
    :param fluor_repr: ``FluorReprType``
    :param plane: ``PLANE_TYPE`` {'coronal', 'sagittal', 'transverse'}
    :param df: if None, give ``ccf_dir`` arg
    :param invert_hemi: if True, then ML_location >= 0 is ipsilateral site. otherwise, is contralateral site
    :return: DataFrame
    """

    if df is None:
        df = _concat_channel(ccf_dir, fluor_repr, plane)

    # pick up acronym with capital
    df = df.filter(pl.col('acronym').str.contains(r'[A-Z]+'))

    df = df.with_columns(
        # exclude words after `area` in the name col
        pl.col('name').str.extract(r'(.+area).*').fill_null(pl.col('name')).alias('abbr'),
        # remove subregion and layer info
        pl.col('acronym').str.extract(r'(CA1|CA2|CA3|[A-Za-z]+).*').fill_null(pl.col('acronym')).alias('acronym_abbr'),
    )

    # add hemisphere info (ml > 0 represent image right site, thus ipsilateral)
    if invert_hemi:
        expr = pl.when(pl.col('ML_location') < 0)
    else:
        expr = pl.when(pl.col('ML_location') >= 0)

    df = df.with_columns(
        expr
        .then(pl.lit('ipsi'))
        .otherwise(pl.lit('contra'))
        .alias('hemi.')
    )

    return df


def supply_overlap_dataframe(df: pl.DataFrame, supply_channel_name: Iterable[Channel]) -> pl.DataFrame:
    """Supply overlap counting in the parsed dataframe.

    ** Only used if excluding the overlap cell while counting the rfp and gfp channel


    :param df: dataframe contain `source` column with `overlap` literal in the cell
    :param supply_channel_name: channel names to be supplied from `overlap`
    """

    ret = [df]
    for name in supply_channel_name:
        s = df.filter(pl.col('source') == 'overlap').with_columns(pl.lit(name).alias('source'))
        ret.append(s)

    return pl.concat(ret)



from __future__ import annotations

import dataclasses
import io
from pathlib import Path

import pandas as pd
import polars as pl

from neuralib.atlas.data import load_structure_tree
from neuralib.io.core import ATLAS_CACHE_DIRECTORY
from neuralib.util.verbose import fprint

__all__ = ['CellAtlas']


@dataclasses.dataclass
class CellAtlas:
    dataframe: pl.DataFrame
    """
    Example::
    
        ┌────────────────────────────────┬────────────────┬───────────┐
        │ Brain region                   ┆ Volumes [mm^3] ┆ n_neurons │
        │ ---                            ┆ ---            ┆ ---       │
        │ str                            ┆ f64            ┆ i64       │
        ╞════════════════════════════════╪════════════════╪═══════════╡
        │ Abducens nucleus               ┆ 0.015281       ┆ 1324      │
        │ Accessory facial motor nucleus ┆ 0.013453       ┆ 497       │
        │ Accessory olfactory bulb       ┆ 0.6880625      ┆ 189608    │
        │ …                              ┆ …              ┆ …         │
        │ Zona incerta                   ┆ 2.157641       ┆ 136765    │
        │ posteromedial visual area      ┆ 1.2225625      ┆ 197643    │
        └────────────────────────────────┴────────────────┴───────────┘
    """

    @classmethod
    def load_from_csv(cls,
                      file: Path | None = None,
                      ignore_cell_types_info: bool = True,
                      ignore_detail_info: bool = True) -> CellAtlas:
        """
        Load/Download the csv file

        :param file: filepath. If None, download from source paper
        :param ignore_cell_types_info: ignore cell types information, only select neuron and volume foreach areas
        :param ignore_detail_info: ignore information in brain subregion
        :return:
        """
        if file is None:
            d = ATLAS_CACHE_DIRECTORY

            if not d.exists():
                d.mkdir(exist_ok=True, parents=True)

            file = d / 'cellatlas.csv'
            if not file.exists():
                cls._request(file)

        df = pl.read_csv(file)

        if ignore_cell_types_info:
            df = df.select('Brain region', 'Neuron [mm^-3]', 'Volumes [mm^3]')

        # total neurons
        df = (
            df.with_columns((pl.col('Neuron [mm^-3]') * pl.col('Volumes [mm^3]')).alias('n_neurons').cast(pl.Int64))
            .drop('Neuron [mm^-3]')
        )

        if ignore_detail_info:
            patterns = (',', '/', r'\(')
            for pt in patterns:
                df = df.filter(~(pl.col('Brain region').str.contains(pt)))

        return CellAtlas(df.sort('Brain region'))

    @classmethod
    def _request(cls, output: Path) -> pl.DataFrame:
        """download from paper source"""
        import requests

        url = 'https://journals.plos.org/ploscompbiol/article/file?type=supplementary&id=10.1371/' \
              'journal.pcbi.1010739.s011'
        resp = requests.get(url)

        if resp.status_code == 200:
            df = pd.read_excel(io.BytesIO(resp.content), sheet_name='Densities BBCAv1')
            pl.from_pandas(df).write_csv(output)
            fprint(f'Download successfully cellatlas csv and save in {output}!', vtype='io')
        else:
            raise RuntimeError('download cellatlas FAIL')

        return df

    @property
    def brain_regions(self) -> list[str]:
        """list of brain regions"""
        return self.dataframe['Brain region'].unique().to_list()

    @classmethod
    def load_sync_allen_structure_tree(cls, force_save: bool = True) -> pl.DataFrame:
        """
        TODO `ProS` not found
        Based on cellatlas dataframe, create a sync used `acronym` header in allen struct_tree (sorted by name)

        **fields**

        ``name``: `Brain region` found in cellatlas

        ``n_neurons``: `n_neurons` from cellatlas dataframe

        ``acronym``: `acronym` found in the structure_tree csv

        :param force_save: create sync file every time in root directory
        :return: sync_dataframe
            example::

                ┌────────────────────────────────┬────────────────┬───────────┬─────────┐
                │ name                           ┆ Volumes [mm^3] ┆ n_neurons ┆ acronym │
                │ ---                            ┆ ---            ┆ ---       ┆ ---     │
                │ str                            ┆ f64            ┆ i64       ┆ str     │
                ╞════════════════════════════════╪════════════════╪═══════════╪═════════╡
                │ Abducens nucleus               ┆ 0.015281       ┆ 1324      ┆ VI      │
                │ Agranular insular area         ┆ 4.901734       ┆ 242362    ┆ AI      │
                │ …                              ┆ …              ┆ …         ┆ …       │
                │ Visual areas                   ┆ 12.957203      ┆ 1297194   ┆ VIS     │
                │ Zona incerta                   ┆ 2.157641       ┆ 136765    ┆ ZI      │
                │ posteromedial visual area      ┆ 1.2225625      ┆ 197643    ┆ VISpm   │
                └────────────────────────────────┴────────────────┴───────────┴─────────┘
        """
        out = ATLAS_CACHE_DIRECTORY / 'cellatlas_allen_sync.csv'

        if not out.exists() or force_save:
            ctlas = cls.load_from_csv()

            allen = load_structure_tree().select('name', 'acronym').sort('name')

            df = (
                ctlas.dataframe
                .rename({'Brain region': 'name'})
                .join(allen, on='name')
            )

            df.write_csv(out)
            fprint(f'SAVE cellatlas sync csv to {out}', vtype='io')

        return pl.read_csv(out)



"""
Swc morphology plot
====================


**Example for CLI usage**

.. code-block:: console

    python -m neuralib.morpho.swc -h


**3D view with radius**

- Press ``Shift+s`` for vedo interactive saving

.. code-block:: console

    python -m neuralib.morpho.swc <SWC_FILE> --radius

**2D view with radius**

.. code-block:: console

    python -m neuralib.morpho.swc <SWC_FILE> --radius --2d


"""

from __future__ import annotations

from pathlib import Path
from typing import NamedTuple, Iterator

import matplotlib.pyplot as plt
import numpy as np
import vedo
from matplotlib.axes import Axes
from matplotlib.patches import Circle
from typing_extensions import Self, overload

from neuralib.argp import AbstractParser, argument
from neuralib.typing import PathLike

__all__ = [
    'SwcNode',
    'SwcFile',
    'plot_swc'
]

Identifier = int
IdentifierName = str

IDENTIFIER_DICT: dict[Identifier, IdentifierName] = {
    0: 'undefined',
    1: 'soma',
    2: 'axon',
    3: 'basal',
    4: 'apical',
    5: 'custom'
}


class SwcNode(NamedTuple):
    n: int
    """node number"""

    identifier: Identifier
    """See IDENTIFIER_DICT"""

    x: float
    """position x"""
    y: float
    """position y"""
    z: float
    """position z"""
    r: float
    """radius"""
    parent: int
    """parent connectivity"""

    @property
    def identifier_name(self) -> IdentifierName:
        return IDENTIFIER_DICT.get(self.identifier, 'custom')

    @property
    def point(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z])

    @property
    def is_undefined(self) -> bool:
        return self.identifier == 0

    @property
    def is_soma(self) -> bool:
        return self.identifier == 1

    @property
    def is_axon(self) -> bool:
        return self.identifier == 2

    @property
    def is_basal_dendrite(self) -> bool:
        return self.identifier == 3

    @property
    def is_apical_dendrite(self) -> bool:
        return self.identifier == 4

    @property
    def is_dendrite(self) -> bool:
        return self.is_basal_dendrite or self.is_apical_dendrite

    @property
    def is_custom(self) -> bool:
        return self.identifier >= 5


class SwcFile:
    node: list[SwcNode]

    def __init__(self, node: list[SwcNode]):
        self.node = node

    @classmethod
    def load(cls, file: PathLike) -> Self:
        """
        :param file: swc filepath
        :return: ``SwcFile``
        """
        node = []
        with Path(file).open('r', encoding='Big5') as f:
            for line in f:
                line = line.strip()

                if len(line) == 0 or line.startswith('#'):
                    continue

                part = line.split()
                n = int(part[0])
                i = int(part[1])
                x = float(part[2])
                y = float(part[3])
                z = float(part[4])
                r = float(part[5])
                p = int(part[6])

                node.append(SwcNode(n, i, x, y, z, r, p))

        return cls(node)

    def __str__(self):
        line = [str(node) for node in self.node]
        return '\n'.join(line)

    @overload
    def __getitem__(self, item: int) -> SwcNode:
        pass

    @overload
    def __getitem__(self, item: IdentifierName) -> Self:
        pass

    def __getitem__(self, item: int | str) -> SwcNode | Self:
        if isinstance(item, int):
            try:
                ret = self.node[item - 1]  # to index
            except IndexError:
                ret = None

            if ret is not None and ret.n == item:
                return ret

            raise KeyError(f'item not found: {item}, might be loss parent connection')

        elif isinstance(item, str):
            if item == 'soma':
                node = [n for n in self.foreach_node() if n.is_soma]
            elif item == 'axon':
                node = [n for n in self.foreach_node() if n.is_axon]
            elif item == 'dendrite':
                node = [n for n in self.foreach_node() if n.is_dendrite]
            elif item == 'basal':
                node = [n for n in self.foreach_node() if n.is_basal_dendrite]
            elif item == 'apical':
                node = [n for n in self.foreach_node() if n.is_apical_dendrite]
            elif item == 'dendrite':
                node = [n for n in self.foreach_node() if n.is_dendrite]
            elif item == 'custom':
                node = [n for n in self.foreach_node() if n.is_custom]
            elif item == 'undefined':
                node = [n for n in self.foreach_node() if n.is_undefined]
            else:
                raise ValueError('')

            return SwcFile(node)

        else:
            raise TypeError(f'item must be int or str: {type(item)}')

    @property
    def points(self) -> np.ndarray:
        return np.array([[n.x, n.y, n.z] for n in self.foreach_node()])

    @property
    def radii(self) -> np.ndarray:
        return np.array([n.r for n in self.foreach_node()])

    @property
    def parents(self) -> np.ndarray:
        return np.array([n.parent for n in self.foreach_node()])

    @property
    def unique_identifier(self) -> list[IdentifierName]:
        idfs = np.unique([n.identifier for n in self.foreach_node()])
        return [
            IDENTIFIER_DICT.get(idf, 'custom')
            for idf in idfs
        ]

    def foreach_identifier(self, as_dict: bool) -> list[Self] | dict[str, Self]:
        if as_dict:
            return {idf: self[idf] for idf in self.unique_identifier}
        else:
            return [self[idf] for idf in self.unique_identifier]

    def foreach_node(self) -> Iterator[SwcNode]:
        for node in self.node:
            yield node

    def foreach_line(self) -> Iterator[tuple[SwcNode, SwcNode]]:
        for node in self.node:
            if node.parent > 0:
                yield node, self[node.parent]


# ============== #
# Plot Functions #
# ============== #

Point3D = tuple[float, float, float]
Point2D = tuple[float, float]

DEFAULT_COLOR: dict[IdentifierName, str] = {
    'soma': 'b',
    'axon': 'r',
    'dendrite': 'k',
    'undefined': 'k',
    'custom': 'k'
}


def projection_2d(p: Point3D) -> Point2D:
    """Default projection function, remove z value.

    :param p: 3d points
    :return: 2d points
    """
    return p[0], p[1]


def smooth_line_radius(ax: Axes,
                       p1: Point2D,
                       p2: Point2D,
                       r1: float,
                       r2: float,
                       num: int = 2,
                       **kwargs):
    """

    :param ax: ``Axes``
    :param p1: Point 1
    :param p2: Point 2
    :param r1: Radius 1
    :param r2: Radius 2
    :param num: Number of segments
    :param kwargs: Additional arguments pass to ``plt.plot()``
    :return:
    """
    px = np.linspace(p1[0], p2[0], num + 1)
    py = np.linspace(p1[1], p2[1], num + 1)
    lw = np.linspace(r1, r2, num)
    for i in range(num):
        ax.plot(px[i:i + 2], py[i:i + 2], lw=lw[i], **kwargs)


def plot_swc(swc: SwcFile,
             radius: bool = True,
             color: dict[str, str] | None = None,
             as_2d: bool = False):
    """
    Plot swc file as 2d

    :param swc: ``SwcFile``
    :param radius: Plot with radius.
    :param color: Color dict. With {identifier name: color coded}
    :param as_2d:
    """
    if color is None:
        color = DEFAULT_COLOR

    if as_2d:
        _plot_swc_2d(swc, radius, color)
    else:
        _plot_swc_3d(swc, radius, color)


def _plot_swc_2d(swc, radius, color):
    fig, ax = plt.subplots()
    for n1, n2 in swc.foreach_line():
        c = color.get(n1.identifier_name, 'k')

        p1 = projection_2d((n1.x, n1.y, n1.z))
        p2 = projection_2d((n2.x, n2.y, n2.z))

        if radius:
            if n2.is_soma:
                ax.add_artist(Circle(p2, n2.r, color=color['soma']))
                if not n1.is_soma:
                    smooth_line_radius(ax, p1, p2, n1.r, n1.r, color=c, solid_capstyle='round')
            else:
                smooth_line_radius(ax, p1, p2, n1.r, n2.r, color=c, solid_capstyle='round')
        else:
            px = p1[0], p2[0]
            py = p1[1], p2[1]
            ax.plot(px, py, color=c, solid_capstyle='round')

    ax.axis('off')
    ax.set_xticklabels([])
    ax.set_yticklabels([])


def _plot_swc_3d(swc: SwcFile,
                 radius,
                 color,
                 spheres_size: float = 3,
                 lw: float = 5):
    plotter = vedo.Plotter()

    axons = []
    axons_line = []
    axons_radii = []

    dendrites = []
    dendrites_line = []
    dendrites_radii = []

    somata = []
    somata_line = []
    somata_radii = []

    other = []
    other_line = []
    other_radii = []

    for i, n in enumerate(swc.foreach_node()):
        if n.parent == -1:
            continue

        r = n.r * spheres_size if radius else 5

        if n.is_axon:
            axons.append([n.x, n.y, n.z])
            axons_line.append([n.parent - 1, i])  # Use parent-child connection for axons
            axons_radii.append(r)
        elif n.is_dendrite:
            dendrites.append([n.x, n.y, n.z])
            dendrites_line.append([n.parent - 1, i])
            dendrites_radii.append(r)
        elif n.is_soma:
            somata.append([n.x, n.y, n.z])
            somata_line.append([n.parent - 1, i])
            somata_radii.append(10)  # fix value

        elif n.is_undefined or n.is_custom:
            other.append([n.x, n.y, n.z])
            other_line.append([n.parent - 1, i])
            other_radii.append(r)

    #
    if 'soma' in swc.unique_identifier:
        soma_spheres = vedo.Spheres(somata, r=somata_radii, c=color['soma'])
        soma_lines = vedo.Lines(swc.points[somata_line], c=color['soma'], lw=lw)
        plotter += soma_spheres
        plotter += soma_lines

    if 'dendrite' in swc.unique_identifier:
        dendrite_spheres = vedo.Spheres(dendrites, r=dendrites_radii, c=color['dendrite'])
        dendrite_lines = vedo.Lines(swc.points[dendrites_line], c=color['dendrite'], lw=lw)
        plotter += dendrite_spheres
        plotter += dendrite_lines

    if 'axon' in swc.unique_identifier:
        axon_spheres = vedo.Spheres(axons, r=axons_radii, c=color['axon'])
        axon_lines = vedo.Lines(swc.points[axons_line], c=color['axon'], lw=lw)
        plotter += axon_spheres
        plotter += axon_lines

    other_spheres = vedo.Spheres(other, r=other_radii, c=color.get('custom', 'k'))
    other_lines = vedo.Lines(swc.points[other_line], c=color.get('custom', 'k'), lw=lw)
    plotter += other_spheres
    plotter += other_lines

    plotter.show()


# ======== #
# Plot CLI #
# ======== #

class SwcPlotOptions(AbstractParser):
    file: str = argument(
        metavar='FILE',
        help='filepath of the swc file'
    )

    radius: bool = argument(
        '-R', '--radius',
        help='Whether plot with radius'
    )

    as_2d: bool = argument(
        '--2d',
        help='Whether plot with 2d, otherwise, plot as 3d'
    )

    def run(self):
        swc = SwcFile.load(self.file)
        plot_swc(swc, radius=self.radius, as_2d=self.as_2d)


if __name__ == '__main__':
    SwcPlotOptions().main()



from __future__ import annotations

from typing import Literal, Sequence

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.offsetbox import AnchoredOffsetbox
from matplotlib.transforms import Transform

from neuralib.typing import ArrayLike, PathLike

__all__ = [
    'AnchoredScaleBar',
    'AxesExtendHelper',
    'insert_latex_equation'
]


class AnchoredScaleBar(AnchoredOffsetbox):
    def __init__(
            self,
            transform: Transform,
            sizex: float = 0,
            sizey: float = 0,
            labelx: str | None = None,
            labely: str | None = None,
            loc: int | str = 4,
            pad: float = 0.1,
            borderpad=0.1,
            sep: int = 2,
            color: str = "black",
            lw: float = 1.5,
            color_txt: str = "black",
            prop=None,
            **kwargs
    ):
        """
        Set of scale bars that match the size of the ticks of the plot.

        Draws a horizontal and/or vertical bar with the size in data coordinates
        of the give axes. A label will be drawn underneath (center-aligned).

        :param transform: Matplotlib Transform
            The coordinate frame (typically axes.transData)
        :param sizex: Width of x bar, in data units. 0 to omit.
        :param sizey: Width of y bar, in data units. 0 to omit.
        :param labelx: Labels for x bars; None to omit
        :param labely: Labels for y bars; None to omit
        :param loc:  Position in containing axes.
        :param pad: Padding, in fraction of the legend font size (or prop).
        :param borderpad:
        :param sep: Separation between labels and bars in points.
        :param color: Bars color
        :param lw:  Bars width
        :param color_txt: color
        :param prop:  Font property.
        :param kwargs: additional arguments passed to base class constructor
        """

        from matplotlib.patches import Rectangle
        from matplotlib.offsetbox import (
            AuxTransformBox,
            VPacker,
            HPacker,
            TextArea,
        )

        bars = AuxTransformBox(transform)

        if sizex:
            bars.add_artist(Rectangle((0, 0), sizex, 0, fc="none", color=color, lw=lw))
        if sizey:
            bars.add_artist(Rectangle((0, 0), 0, sizey, fc="none", color=color, lw=lw))

        if sizex and labelx:
            bars = VPacker(children=[bars, TextArea(labelx, minimumdescent=False, textprops=dict(color=color_txt))],
                           align="center",
                           pad=0,
                           sep=sep)
        if sizey and labely:
            bars = HPacker(children=[TextArea(labely, textprops=dict(color=color_txt)), bars],
                           align="center",
                           pad=0,
                           sep=sep)

        AnchoredOffsetbox.__init__(
            self,
            loc,
            pad=pad,
            borderpad=borderpad,
            child=bars,
            prop=prop,
            frameon=False,
            **kwargs
        )


# ========================= #

class AxesExtendHelper:
    """
    Add a child inset Axes to this existing Axes

    Example of add hist::

        >>> from neuralib.plot import plot_figure

        >>> x = np.random.sample(10)
        >>> y = np.random.sample(10)
        >>> with plot_figure(None) as ax:
        ...     ax.plot(x, y, 'k.')
        ...     helper = AxesExtendHelper(ax)
        ...     helper.xhist(x, bins=10)
        ...     helper.yhist(y, bins=10)


    Example of add bar::

        >>> from neuralib.plot import plot_figure
        >>> img = np.random.sample((10, 10))
        >>> with plot_figure(None) as ax:
        ...     ax.imshow(img)
        ...     helper = AxesExtendHelper(ax)
        ...     x = y = np.arange(10)
        ...     helper.xbar(x, np.mean(img, axis=0), align='center')
        ...     helper.ybar(y, np.mean(img, axis=1), align='center')

    """
    ax_x: Axes | None
    ax_y: Axes | None

    def __init__(self,
                 ax: Axes,
                 mode: Literal['both', 'x', 'y'] = 'both'):
        """

        :param ax: :class:`matplotlib.axes.Axes`
        :param mode: extended axis {'both', 'x', 'y'}. default is to add `both`
        """
        ax.set(aspect=1)
        if mode in ('both', 'x'):
            self.ax_x = ax.inset_axes((0, 1.05, 1, 0.25), sharex=ax)
        else:
            self.ax_x = None

        if mode in ('both', 'y'):
            self.ax_y = ax.inset_axes((1.05, 0, 0.25, 1), sharey=ax)
        else:
            self.ax_y = None

        self.ax = ax

    def xhist(self, values: ArrayLike,
              bins: int | Sequence[float] | str | None,
              **kwargs):
        """x axis histogram

        :param values: histogram x axis
        :param bins: number of bins
        :param kwargs: additional arguments passed to ``Axes.hist()``
        """
        if self.ax_x is not None:
            self.ax_x.hist(values, bins, **kwargs)

    def yhist(self, values: ArrayLike,
              bins: int | Sequence[float] | str | None,
              **kwargs):
        """y axis histogram

        :param values: histogram x axis
        :param bins: number of bins
        :param kwargs: additional arguments passed to ``Axes.hist()``
        """
        if self.ax_y is not None:
            self.ax_y.hist(values, bins, orientation='horizontal', **kwargs)

    def xbar(self, x: ArrayLike,
             height: np.ndarray,
             width: float | np.ndarray | None = None,
             **kwargs):
        """x axis bar

        :param x: x axis
        :param height: bar height
        :param width: bar width
        :param kwargs: additional arguments passed to ``Axes.bar()``
        """
        default_kw = {
            'color': 'grey',
            'edgecolor': 'black',
            'align': 'edge'
        }

        kw = default_kw | kwargs

        if width is None:
            width = np.diff(x)[0]

        if self.ax_x is not None:
            self.ax_x.bar(x, height, width=width, **kw)

            self.ax_x.tick_params(axis="x", labelbottom=False)

    def ybar(self, y: ArrayLike,
             width: np.ndarray,
             height: float | np.ndarray | None = None,
             **kwargs):
        """y axis bar

        :param y: y axis
        :param height: bar height
        :param width: bar width
        :param kwargs: additional arguments passed to ``Axes.bar()``
        """
        default_kw = {
            'color': 'grey',
            'edgecolor': 'black',
            'align': 'edge'
        }

        kw = default_kw | kwargs

        if height is None:
            height = np.diff(y)[0]

        if self.ax_y is not None:
            self.ax_y.barh(y, width, height=height, **kw)

            self.ax_y.tick_params(axis="y", labelleft=False)


# ========================= #

def insert_latex_equation(ax: Axes,
                          tex: str,
                          output: PathLike | None = None):
    """Plot figure with latex expression, and save as image

    :param ax: :class:`matplotlib.axes.Axes`
    :param tex: latex string
    :param output: output of the fig, if None then save the figure
    """
    plt.rcParams['text.usetex'] = True

    ax.set_title(tex)
    if output is not None:
        from matplotlib.mathtext import math_to_image
        math_to_image(tex, output, dpi=300)



from typing import Generic, TypeVar, Callable

from bokeh.document import Document

__all__ = ['TimeoutUpdateValue']

T = TypeVar('T')
missing = object()


class TimeoutUpdateValue(Generic[T]):
    def __init__(self,
                 document: Document,
                 callback: Callable[[T], None],
                 delay=1000):
        self.__document = document
        self.__value = missing
        self.__callback = callback
        self.delay = delay

    def update(self, value: T):
        old_value = self.__value
        self.__value = value
        if old_value is missing:
            self.__document.add_timeout_callback(self.callback, self.delay)

    def callback(self):
        value = self.__value
        self.__value = missing
        if value is not missing:
            self.__callback(value)



"""
Cellpose
=========

API Mode
---------


**Example of run an image segmentation in 2D mode and visualize using napari**

.. code-block:: console

    python -m neuralib.segmentation.cellpose.run_api -F <IMAGE_FILE> --napari

- It produces an output ``<IMAGE_FILE>_seg.npy`` in the same directory, If the output exists, the napari will directly use the output. If need run again and rewrite the output, use ``--force-eval`` option


**Example of run an image segmentation in 2D mode and visualize using Cellpose GUI**

.. code-block:: console

    python -m neuralib.segmentation.cellpose.run_api -F <IMAGE_FILE> --cpose



**Example of run batch image segmentation in 2D mode from a directory (.tif files)**

.. code-block:: console

    python -m neuralib.segmentation.cellpose.run_2d -D <DIRECTORY> --suffix .tif

- It produces multiple ``*_seg.npy`` in the same directory.


**See help using** ``-h`` **option**

.. code-block:: console

    python -m neuralib.segmentation.cellpose.run_api -h




Subprocess Mode
-----------------


**Example of run an image segmentation in subprocess call**

.. code-block:: console

    python -m neuralib.segmentation.cellpose.run_subproc -F <IMAGE_FILE>

- It produces an output ``<IMAGE_FILE>_seg.npy`` in the same directory, If the output exists, the napari will directly use the output. If need run again and rewrite the output, use ``--force-eval`` option



**Example of run batch image segmentation in subprocess call**

.. code-block:: console

    python -m neuralib.segmentation.cellpose.run_subproc -D <DIRECTORY>

- It produces multiple ``*.npy`` in the same directory.


**See help using** ``-h`` **option**

.. code-block:: console

    python -m neuralib.segmentation.cellpose.run_subproc -h


"""



from __future__ import annotations

import numpy as np

from .norm import get_percentile_value

__all__ = ['grey2rgb']


def grey2rgb(im: np.ndarray, rgb: int | tuple[int, ...]) -> np.ndarray:
    """
    Converts a grayscale image to an RGB image by assigning intensity values
    to specified color channels.

    :param im: Grey scale image (2d array)
    :param rgb: {0, 1, 2} indicates rgb, tuple represent merge color
    :return: RGB image. `Array[float, [H, W, 3]]`
    """

    if im.ndim != 2:
        raise ValueError('input should be an grayscale 2d array')

    lb, ub = get_percentile_value(im, perc_interval=(10, 100))
    if ub == lb:
        raise ValueError("Cannot normalize with zero range in image.")

    out = np.zeros((*im.shape, 3), float)
    if isinstance(rgb, int):
        out[:, :, rgb] = (im - lb) / ub
    else:
        for i in rgb:
            out[:, :, i] = (im - lb) / ub

    return np.clip(out, 0, 1)



"""
SequenceLabeller
===================

Simple CV2-based viewer/labeller GUI for image sequences

Use Cases:

- viewing the image sequences

- label each image and save as csv dataframe (human-eval for population neurons activity profile)


Load sequences from a directory
-----------------------------------

- Use CLI mode

See help::

    python -m neuralib.imglib.labeller -h

Example::

    python neuralib.imglib.labeller -D <DIR>


- Use API call

.. code-block:: python

    from neuralib.imglib.labeller import SequenceLabeller

    directory = ...
    labeller = SequenceLabeller.load_from_dir(directory)
    labeller.main()


Load sequences from sequences array
-------------------------------------

.. code-block:: python

    from neuralib.imglib.labeller import SequenceLabeller

    arr = ...   # numpy array with (F, H, W, <3>)
    labeller = SequenceLabeller.load_sequences(arr)
    labeller.main()

"""

from __future__ import annotations

import logging
import re
import sys
from pathlib import Path
from typing import ClassVar, Callable, Any, Literal, TypedDict

import attrs
import cv2
import numpy as np
import polars as pl
from polars.testing import assert_frame_equal
from tifffile import tifffile
from tqdm import tqdm
from typing_extensions import Self

from neuralib.io import csv_header
from neuralib.typing import PathLike
from neuralib.util.utils import keys_with_value
from neuralib.util.verbose import fprint

__all__ = ['SequenceLabeller']

Logger = logging.getLogger(__name__)


# ================ #
# KeyBoard Mapping #
# ================ #

class KeyMapping(TypedDict, total=False):
    """For controlling the keyboard in different OS"""
    escape: int
    backspace: int
    # space: int  # labeller specific printable for notes
    enter: int

    left_square_bracket: int  # [
    right_square_bracket: int  # ]

    left: int
    right: int
    up: int
    down: int

    plus: int  # +
    minus: int  # -
    equal: int  # =


#
COMMON_KEYMAPPING: KeyMapping = {
    'escape': 27,
    # 'space': 32,
    'enter': 13,

    'left_square_bracket': 91,
    'right_square_bracket': 93,

    # ord
    'plus': ord('+'),
    'minus': ord('-'),
    'equal': ord('=')
}

WIN_KEYMAPPING: KeyMapping = {
    **COMMON_KEYMAPPING,
    'backspace': 8,
    'left': 2424832,
    'right': 2555904,
    'up': 2490368,
    'down': 2621440
}

MAC_KEYMAPPING: KeyMapping = {
    **COMMON_KEYMAPPING,
    'backspace': 127,
    'left': 2,
    'right': 3,
    'up': 0,
    'down': 1

}

LINUX_KEYMAPPING: KeyMapping = {
    **COMMON_KEYMAPPING,
    'backspace': 8,
    'left': 81,
    'right': 83,
    'up': 82,
    'down': 84
}


def get_keymapping() -> KeyMapping:
    p = sys.platform
    if p in ('linux', 'linux2'):
        return LINUX_KEYMAPPING
    elif p == 'darwin':
        return MAC_KEYMAPPING
    elif p == 'win32':
        return WIN_KEYMAPPING


@attrs.define
class FrameInfo:
    filename: str
    """name of the an image/frame"""

    image: np.ndarray
    """`Array[uint, [H, W]|[H, W, 3])`"""

    notes: str | None = attrs.field(default=None)
    """notes for the image"""

    @property
    def itype(self) -> Literal['gray', 'rgb']:
        """image color type"""
        if self.image.ndim == 2:
            return 'gray'
        elif self.image.ndim == 3:
            return 'rgb'
        else:
            raise TypeError('')

    @property
    def text_color(self) -> int | tuple[int, int, int]:
        if self.itype == 'gray':
            return 2 ** 16 - 1
        elif self.itype == 'rgb':
            return 0, 0, 255
        else:
            raise RuntimeError('')

    @property
    def height(self) -> int:
        return self.image.shape[0]

    @property
    def width(self) -> int:
        return self.image.shape[1]


class CloseSaveInterrupt(KeyboardInterrupt):
    """write & quiet triggered"""

    def __init__(self, mode: Literal[':wq', ':q', ':q!']):
        self.mode = mode


class SequenceLabeller:
    window_title: ClassVar[str] = 'SeqLabeller'

    def __init__(self, seqs_info: list[FrameInfo],
                 output: PathLike | None = None):
        self.seqs_info = seqs_info
        self.output = output  # for notes

        self.message_queue: list[str] = []
        self.buffer = ''  # input buffer
        self._frame_index = 0

    def __len__(self) -> int:
        return len(self.seqs_info)

    @classmethod
    def load_sequences(cls, seqs: np.ndarray | list[np.ndarray],
                       filenames: list[str] | None = None,
                       output: PathLike | None = None) -> Self:
        """

        :param seqs:
        :param filenames:
        :param output:
        :return:
        """
        if isinstance(seqs, np.ndarray):
            seqs = list(seqs)

        n_frames = len(seqs)

        if filenames is None:
            filenames = np.arange(n_frames).astype(str)

        seqs_info = [FrameInfo(filenames[i], seqs[i], None) for i in range(n_frames)]

        return SequenceLabeller(seqs_info, Path(output) if output is not None else None)

    @classmethod
    def load_from_dir(cls, directory: PathLike,
                      file_suffix: str = '.tif',
                      sort_func: Callable[[Path], Any] | None = None,
                      single_frame_per_file: bool = True,
                      output: PathLike | None = None) -> Self:
        """

        :param directory: directory contain image sequences
        :param file_suffix: sequence file suffix
        :param sort_func: sorted function with signature `(filename:Path) -> Comparable`
        :param single_frame_per_file:
        :param output:
        :return:
        """
        if not Path(directory).is_dir():
            raise NotADirectoryError(f'{directory}')

        files = sorted(list(directory.glob(f'*{file_suffix}')), key=sort_func)

        if len(files) == 0:
            raise FileNotFoundError('')
        else:
            fprint(f'LOAD image sequence: {len(files)} files', vtype='io')

        seqs = []
        for f in tqdm(files, unit='file', ncols=80):

            if file_suffix == '.pdf':
                from neuralib.imglib.io import read_pdf
                img = read_pdf(f, dpi=200)
                seqs.append(img)
            else:
                if single_frame_per_file:
                    img = cv2.imread(str(f))
                    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
                    seqs.append(img)
                elif not single_frame_per_file and file_suffix in ('.tif', '.tiff'):
                    seqs.append(tifffile.imread(str(f)))
                else:
                    raise NotImplementedError('')

        filenames = [it.stem for it in files]
        seqs_info = [FrameInfo(filenames[i], seqs[i], None) for i in range(len(seqs))]

        return SequenceLabeller(seqs_info, Path(output) if output is not None else None)

    @property
    def n_frames(self) -> int:
        """aka. number of images"""
        return len(self.seqs_info)

    @property
    def current_frame_index(self) -> int:
        return self._frame_index

    @current_frame_index.setter
    def current_frame_index(self, value: int):
        n = len(self.seqs_info)
        self._frame_index = (value + n) % n
        self.message_queue = []

        info = self.seqs_info[self._frame_index]
        if info.filename is not None:
            self.enqueue_message(f'{info.filename}')
        if (note := self.read_note()) is not None:
            self.enqueue_message(note)

    @property
    def text_color(self) -> float | tuple[int, int, int]:
        return self.seqs_info[self.current_frame_index].text_color

    # ===== #
    # Notes #
    # ===== #

    def save_note(self):
        """save image-related notes to file"""
        from datetime import datetime

        fields = ['filename', 'notes', 'datetime']
        t = datetime.now().replace(second=0, microsecond=0).strftime("%Y-%m-%d %H:%M")  # TODO to frame dependent
        with csv_header(self.output, fields, quotes_header='notes') as csv:
            for seq in self.seqs_info:
                csv(str(seq.filename), seq.notes, t)

    _prev_note: pl.DataFrame | None = None  # for checking changes

    def load_note(self):
        """read image-related notes from file"""
        from neuralib.util.verbose import printdf

        df = self._prev_note = pl.read_csv(self.output, schema_overrides={'filename': pl.Utf8})
        printdf(df)
        for i, info in enumerate(self.seqs_info):
            note = df.filter(pl.col('filename') == info.filename)['notes'].item()
            self.seqs_info[i].notes = note

    def write_note(self, note: str, *, append_mode: bool = False):
        if append_mode:
            prev = self.seqs_info[self.current_frame_index].notes
            self.seqs_info[self.current_frame_index].notes = prev + ';' + note
        else:
            self.seqs_info[self.current_frame_index].notes = note

        self.current_frame_index = self.current_frame_index  # trigger enqueue_message

        if self.output is None:
            self.enqueue_message('specify output first for writing notes!')

    def read_note(self) -> str | None:
        return self.seqs_info[self._frame_index].notes

    def clear_note(self):
        self.seqs_info[self.current_frame_index].notes = None

    def check_note_changes(self) -> bool:
        """True if any changes in notes"""
        if self._prev_note is None:
            if any([seq.notes is not None for seq in self.seqs_info]):
                return True
        else:
            prev_note = self._prev_note.select('filename', 'notes')
            cur_note = pl.DataFrame([
                [str(seq.filename), seq.notes]
                for seq in self.seqs_info
            ], schema=['filename', 'notes'], orient='row')

            try:
                assert_frame_equal(prev_note, cur_note)
            except AssertionError:
                return True

        return False

    # ============= #
    # Key & Command #
    # ============= #

    def goto_begin(self):
        self.current_frame_index = 0

    def goto_end(self):
        self.current_frame_index = self.n_frames - 1

    def go_to(self, i: int):
        if i > len(self) or i < 0:
            self.enqueue_message(f'invalid sequence index: {i}')
            return
        self.current_frame_index = i

    def handle_keycode(self, k: int):
        mapping = get_keymapping()
        ret = self._handle_keymapping(mapping, k)
        if ret is not None:  # printable
            self.buffer += chr(k)

    def _handle_keymapping(self, mapping: KeyMapping, value: int) -> int | None:
        """
        Handling the keyboard mapping

        :param mapping:
        :param value:
        :return: int value if cannot find key in keymapping, otherwise return None
        """
        try:
            ret = keys_with_value(mapping, value, to_item=True)
        except (KeyError, ValueError):
            return value
        else:
            if ret == 'left':
                self.current_frame_index -= 1
            elif ret == 'right':
                self.current_frame_index += 1
            elif ret == 'left_square_bracket':
                self.current_frame_index += 10
            elif ret == 'right_square_bracket':
                self.current_frame_index -= 10
            elif ret == 'backspace':
                if len(self.buffer) > 0:
                    self.buffer = self.buffer[:-1]
            elif ret == 'enter':  # handle command in buffer
                command = self._proc_image_command = self.buffer
                self.buffer = ''
                try:
                    self.handle_command(command)
                except KeyboardInterrupt:
                    raise
                except BaseException as e:
                    self.enqueue_message(f'command "{command}" {type(e).__name__}: {e}')
            elif ret == 'escape':
                self.buffer = ''

    def handle_command(self, command: str):
        Logger.debug(f'command: {command}')

        if command == ':h':
            self.enqueue_message(':h : print this document')
            self.enqueue_message(':q : quit (unable if not save the changes)')
            self.enqueue_message(':q! : quit (without save)')
            self.enqueue_message(':wq : save notes and quit')
            self.enqueue_message(':c : clear current note')
            self.enqueue_message(':i : print current file index')
            self.enqueue_message(':N : goto N-th image')
            self.enqueue_message('+message : append note')
            self.enqueue_message('message : (replace) note')

        elif command == ':c':
            self.enqueue_message(f'clear notes: {self.seqs_info[self.current_frame_index].notes}')
            self.clear_note()
        elif command == ':i':
            self.enqueue_message(f'current file index: {self.current_frame_index}')
        elif re.match(r'^:(\d)', command):
            match = re.search(r'^:(\d)', command)
            self.go_to(int(match.group(1)))
        elif command.startswith('+'):
            self.write_note(command[1:], append_mode=True)
        elif not command.startswith(':'):
            self.write_note(command)
        elif command in (':wq', ':q', ':q!'):
            raise CloseSaveInterrupt(command)
        else:
            raise RuntimeError(f'unknown command : "{command}"')

    # ============ #
    # Msg / Buffer #
    # ============ #

    def enqueue_message(self, text: str):
        self.message_queue.append(text)

    def _show_queued_message(self, image: np.ndarray):
        """drawing enqueued message"""
        y = 70
        s = 30
        i = 0
        while i < len(self.message_queue):
            m = self.message_queue[i]
            cv2.putText(image, m, (10, y), cv2.FONT_HERSHEY_SIMPLEX, 1, self.text_color, 2, cv2.LINE_AA)
            i += 1
            y += s

    def _show_buffer(self, image):
        """drawing input buffer content"""
        cv2.putText(image, self.buffer, (10, 30), cv2.FONT_HERSHEY_SIMPLEX, 1, self.text_color, 2, cv2.LINE_AA)

    # ========= #
    # Main Loop #
    # ========= #

    def main(self):
        """main loop for the GUI"""
        cv2.namedWindow(self.window_title, cv2.WINDOW_GUI_NORMAL)

        if self.output is not None:
            if self.output.exists():
                self.load_note()

        try:
            while True:
                try:
                    while True:
                        self._loop()
                except CloseSaveInterrupt as e:
                    if e.mode == ':wq':
                        if self.output is not None:
                            self.save_note()
                            fprint(f'SAVE csv -> {str(self.output)}!', vtype='io')
                        break
                    elif e.mode == ':q!':
                        break
                    elif e.mode == ':q':
                        if self.check_note_changes():
                            self.enqueue_message('please save the note using :wq, or force quit using :q!')
                            continue
                        else:
                            break
        finally:
            cv2.destroyWindow(self.window_title)

    def _loop(self):
        #
        try:
            info = self.seqs_info[self.current_frame_index]
        except IndexError:
            pass
        else:
            image = info.image.copy()

            if len(self.buffer):
                self._show_buffer(image)
            self._show_queued_message(image)

            cv2.imshow(self.window_title, image)

        #
        if sys.platform in ('darwin', 'linux', 'linux2'):
            k = cv2.waitKey(1)
        elif sys.platform == 'win32':
            k = cv2.waitKeyEx(1)
        else:
            raise RuntimeError('')

        if k >= 0:
            self.handle_keycode(k)


# ============= #
# Main Argparse #
# ============= #

def main():
    import argparse

    ap = argparse.ArgumentParser(description='run the sequences labeller')

    ap.add_argument('-D', '--dir', type=Path, required=True, help='path with image sequences', dest='directory')
    ap.add_argument('--suffix', choices=('.pdf', '.tif', '.tiff'), default='.pdf', help='image sequence suffix')
    ap.add_argument('-O', '--output', type=Path, default=None, help='csv output for note')

    opt = ap.parse_args()

    labeller = SequenceLabeller.load_from_dir(opt.directory,
                                              file_suffix=opt.suffix,
                                              output=opt.output)

    labeller.main()


if __name__ == '__main__':
    main()



"""
OASIS
========

**Fast online deconvolution of calcium imaging dat**


Method source
--------------

- `<https://github.com/j-friedrich/OASIS>`_

- `<https://journals.plos.org/ploscompbiol/article?id=10.1371/journal.pcbi.1005423>`_


This script adapted from ``suite2p.suite2p.extraction.dcnv``
(Copyright © 2023 Howard Hughes Medical Institute, Authored by Carsen Stringer and Marius Pachitariu)

.. seealso:: `<https://github.com/MouseLand/suite2p/blob/main/suite2p/extraction/dcnv.py>`_


**Example of usage**

.. code-block:: python

    from neuralib.imaging.spikes.oasis import oasis_dcnv

    # 2D dF/F array. Array[float, [nNeurons, nFrames]] or Array[float, nFrames]
    dff = ...

    tau = 1.5  # time constant of the calcium indicator (ms)
    fs = 30  # sampling frequency of the calcium imaging data (hz)
    spks = oasis_dcnv(dff, tau, fs)


"""
from __future__ import annotations

import numpy as np
from numba import njit, prange

__all__ = ['oasis_dcnv',
           'oasis_matrix']


def oasis_dcnv(dff: np.ndarray,
               tau: float,
               fs: float,
               batch_size: int = 300) -> np.ndarray:
    """
    Computes non-negative deconvolution (no sparsity constraints)

    :param dff: The observed calcium fluorescence trace in 2D numpy array(multiple neurons)
        or 1D numpy array(single cell). `Array[float, [N, F]|F]`
    :param tau: The time constant of the calcium indicator in ms
    :param fs: The sampling frequency of the calcium imaging data in hz
    :param batch_size: number of frames processed per batch
    :return: Deconvolved fluorescence. `Array[float, [N,F]|F]`
    """
    dff = dff.astype(np.float32)
    if dff.ndim == 1:  # single neurons
        n_neurons = 1
        dff = np.expand_dims(dff, 0)
    else:
        n_neurons = dff.shape[0]

    n_frames = dff.shape[1]

    if batch_size > n_frames:
        batch_size = n_frames

    ret = np.zeros(dff.shape, dtype=np.float32)
    for i in range(0, n_neurons, batch_size):
        f = dff[i:i + batch_size]
        v = np.zeros(dff.shape, dtype=np.float32)
        w = np.zeros(dff.shape, dtype=np.float32)
        t = np.zeros(dff.shape, dtype=np.int64)
        ll = np.zeros(dff.shape, dtype=np.float32)
        s = np.zeros(dff.shape, dtype=np.float32)
        oasis_matrix(f, v, w, t, ll, s, tau, fs)
        ret[i:i + batch_size] = s

    if n_neurons == 1:
        return ret[0]
    else:
        return ret


@njit(parallel=True, cache=True)
def oasis_matrix(
        dff: np.ndarray,
        v: np.ndarray,
        w: np.ndarray,
        t: np.ndarray,
        ll: np.ndarray,
        s: np.ndarray,
        tau: float,
        fs: float
):
    """ Performs spike deconvolution for a single neuron's calcium imaging trace using a greedy method

    Iterates through each time point in the observed fluorescence signal ``dff`` and enforces
    a non-negative and non-increasing constraint on the estimated signal.
    When a violation of this constraint is detected, the function merges the current segment with the previous one,
    updating the estimated signal accordingly. Finally, it computes the inferred spikes by
    calculating the difference between successive segments in the deconvolved signal.

    :param dff:  The observed calcium fluorescence trace. `Array[float, [N, F]|F]`
    :param v: A 1D array that will store the estimated deconvolved signal.
    :param w: A 1D array that tracks the weights for merging steps
    :param t: A 1D array that stores the indices of time steps
    :param ll: A 1D array that tracks the weights for merging steps.
    :param s: A 1D array that will store the inferred spikes
    :param tau: The time constant of the calcium indicator
    :param fs: The sampling frequency of the calcium imaging data

    """
    n_neurons = dff.shape[0]
    for n in prange(n_neurons):
        _oasis_trace(dff[n], v[n], w[n], t[n], ll[n], s[n], tau, fs)


@njit(cache=True)
def _oasis_trace(dff, v, w, t, ll, s, tau, fs):
    """single neurons spike deconvolution"""
    nframes = len(dff)
    g = -1. / (tau * fs)

    i = 0
    ip = 0

    while i < nframes:
        v[ip] = dff[i]
        w[ip] = 1
        t[ip] = i
        ll[ip] = 1

        while ip > 0:
            if v[ip - 1] * np.exp(g * ll[ip - 1]) > v[ip]:
                # violation of the constraint means merging pools
                f1 = np.exp(g * ll[ip - 1])
                f2 = np.exp(2 * g * ll[ip - 1])
                wnew = w[ip - 1] + w[ip] * f2
                v[ip - 1] = (v[ip - 1] * w[ip - 1] + v[ip] * w[ip] * f1) / wnew
                w[ip - 1] = wnew
                ll[ip - 1] = ll[ip - 1] + ll[ip]
                ip -= 1
            else:
                break
        i += 1
        ip += 1

    s[t[1:ip]] = v[1:ip] - v[:ip - 1] * np.exp(g * ll[:ip - 1])



from typing import TypeVar, Callable, Union

__all__ = [
    'try_int_type',
    'try_float_type',
    'int_tuple_type',
    'str_tuple_type',
    'float_tuple_type',
    'tuple_type',
    'union_type',
    'dict_type',
    'slice_type'
]

T = TypeVar('T')


def tuple_type(value_type: Callable[[str], T]):
    def _type(arg: str) -> tuple[T, ...]:
        return tuple(map(value_type, arg.split(',')))

    return _type


str_tuple_type = tuple_type(str)
int_tuple_type = tuple_type(int)
float_tuple_type = tuple_type(float)


def union_type(*t):
    def _type(arg: str):
        for _t in t:
            try:
                return _t(arg)
            except (TypeError, ValueError):
                pass
        raise TypeError

    return _type


def dict_type(default: dict[str, T], value_type: Callable[[str], T] = None):
    """Dict arg value.

    :param default: default dict content
    :param value_type: type of dict value
    :return: type converter
    """
    if default is None:
        default = {}

    def _type(arg: str) -> dict[str, T]:
        if ':' in arg:
            i = arg.index(':')
            value = arg[i + 1:]
            if value_type is not None:
                value = value_type(value)
            default[arg[:i]] = value
        elif '=' in arg:
            i = arg.index('=')
            value = arg[i + 1:]
            if value_type is not None:
                value = value_type(value)
            default[arg[:i]] = value
        elif value_type is None:
            default[arg] = None
        else:
            default[arg] = value_type("")
        return default

    return _type


def slice_type(arg: str) -> slice:
    i = arg.index(':')
    v1 = int(arg[:i])
    v2 = int(arg[i + 1:])
    return slice(v1, v2)


def try_int_type(arg: str) -> Union[int, str, None]:
    """for argparse (i.e., plane_index)"""
    if len(arg) == 0:
        return None
    try:
        return int(arg)
    except ValueError:
        return arg


def try_float_type(arg: str) -> Union[float, str, None]:
    if len(arg) == 0:
        return
    try:
        return float(arg)
    except ValueError:
        return arg



from __future__ import annotations

from pathlib import Path
from pprint import pprint
from typing import Literal, NamedTuple, Iterable

import numpy as np
import pandas as pd
import polars as pl

from neuralib.atlas.data import load_structure_tree
from neuralib.atlas.map import NUM_MERGE_LAYER
from neuralib.atlas.typing import Source, HEMISPHERE_TYPE
from neuralib.typing import DataFrame
from neuralib.util.verbose import fprint

__all__ = [
    'ALLEN_CCF_10um_BREGMA',
    'PLANE_TYPE',
    #
    'SourceCoordinates',
    'iter_source_coordinates',
    'get_margin_merge_level',
    'roi_points_converter',
    'create_allen_structure_dict'
]

# allen CCF 10um volume coordinates, refer to allenCCF/Browsing Functions/allenCCFbregma.m
ALLEN_CCF_10um_BREGMA = np.array([540, 0, 570])  # AP, DV, LR

PLANE_TYPE = Literal['coronal', 'sagittal', 'transverse']


class SourceCoordinates(NamedTuple):
    source: Source
    coordinates: np.ndarray
    """(N, 3) with ap, dv, ml"""

    axes_repr: tuple[str, str, str] = ('ap', 'dv', 'ml')

    @property
    def ap(self) -> np.ndarray:
        return self.coordinates[:, 0]

    @property
    def dv(self) -> np.ndarray:
        return self.coordinates[:, 1]

    @property
    def ml(self) -> np.ndarray:
        return self.coordinates[:, 2]


def iter_source_coordinates(
        file: Path,
        *,
        only_areas: list[str] | str | None = None,
        region_col: str | None = None,
        hemisphere: HEMISPHERE_TYPE = 'both',
        to_brainrender: bool = True,
        to_um: bool = True,
        ret_order: tuple[Source, ...] | None = ('pRSC', 'aRSC', 'overlap')
) -> Iterable[SourceCoordinates]:
    """Load allen ccf roi output (merged different color channels).

    :param file: parsed csv file after
    :param only_areas: only show rois in region(s)
    :param region_col: if None, auto infer, and check the lowest merge level contain all the regions specified
    :param hemisphere: which brain hemisphere
    :param to_brainrender: convert the coordinates to brain render
    :param to_um: unit to um
    :param ret_order: whether specify the source generator order
    :return: Iterable of :class:`SourceCoordinates`
    """
    df = pl.read_csv(file)
    #
    if only_areas is not None and len(only_areas) != 0:
        if isinstance(only_areas, str):
            only_areas = [only_areas]

        if region_col is None:
            region_col = get_margin_merge_level(df, only_areas, 'lowest')

        df = df.filter(pl.col(region_col).is_in(only_areas))
        fprint(f'using {region_col} for {only_areas}')

        if df.is_empty():
            raise RuntimeError('check lowest merge level')
    #
    if hemisphere != 'both':
        df = df.filter(pl.col('hemi.') == hemisphere)

    #
    points = roi_points_converter(df, to_brainrender=to_brainrender, to_um=to_um)

    if ret_order is None:
        ret_order = df['source'].unique()

    for src in ret_order:
        mask = df.select(pl.col('source') == src).to_numpy()[:, 0]
        yield SourceCoordinates(src, points[mask])


def get_margin_merge_level(df: pl.DataFrame,
                           areas: list[str] | str,
                           margin: Literal['lowest', 'highest']) -> str:
    """Get the lowest or highest merge level (i.e., parsed_csv) containing all the regions

    :param df: parsed csv
    :param areas: an area or a list of areas
    :param margin: get the either lowest of highest merge level for a given area
    :return: col name if parsed csv
    """
    if not isinstance(areas, (tuple, list)):
        areas = [areas]

    eval_merge = [f'merge_ac_{i}' for i in range(NUM_MERGE_LAYER)]

    if margin == 'lowest':
        level = eval_merge
    elif margin == 'highest':
        level = reversed(eval_merge)
    else:
        raise ValueError('')

    for lv in level:
        ls = df[lv]
        if np.all([a in ls for a in areas]):
            return lv

    raise ValueError(f'{areas} not found')


def roi_points_converter(dat: DataFrame | np.ndarray,
                         to_brainrender: bool = True,
                         to_um: bool = True) -> np.ndarray:
    """
    convert coordinates of `allenccf` roi points from parsed dataframe

    :param dat: Dataframe with 'AP_location', 'DV_location', 'ML_location' headers.
            Or numpy array with `Array[float, [N, 3]]` or `Array[float, 3]`
    :param to_brainrender: coordinates to `brainrender`
    :param to_um: unit to um
    :return: `Array[float, [N, 3]]`, N: number of roi; 3: AP, DV, ML
    """
    if isinstance(dat, pd.DataFrame):
        dat = pl.from_pandas(dat)
        points = dat.select('AP_location', 'DV_location', 'ML_location').to_numpy()
    elif isinstance(dat, pl.DataFrame):
        points = dat.select('AP_location', 'DV_location', 'ML_location').to_numpy()
    elif isinstance(dat, np.ndarray):
        if dat.ndim == 1:
            dat = np.expand_dims(dat, axis=0)

        if dat.ndim == 2 and dat.shape[1] != 3:
            raise ValueError('')

        points = dat
    else:
        raise TypeError('')

    #
    if to_um:
        points *= 1000  # um

    if to_brainrender:
        points[:, 0] /= -1  # increment toward posterior
        points[:, 2] /= -1  # increment toward left hemisphere
        bregma = ALLEN_CCF_10um_BREGMA * 10  # pixel to um
        points += bregma  # roi relative to bregma

    return points


def create_allen_structure_dict(verbose=False) -> dict[str, str]:
    """
    Get the acronym/name pairing from structure_tree.csv

    :return: key: acronym; value: full name
    """
    tree = load_structure_tree()
    tree = tree.select('name', 'acronym').sort('name')

    ret = {
        acry: name
        for name, acry in tree.iter_rows()
    }
    if verbose:
        pprint(ret)

    return ret



from __future__ import annotations

from typing import final, Literal

import attrs
import cv2
import numpy as np
from typing_extensions import Self

from neuralib.typing import PathLike

__all__ = [
    'ImageProcFactory',
    'IMAGE_CHANNEL_TYPE',
    #
    'recover_overexposure',
]

IMAGE_CHANNEL_TYPE = Literal['red', 'green', 'blue', 'r', 'g', 'b']


@final
@attrs.define(repr=False)
class ImageProcFactory:
    """Factory for basic imaging processing"""

    image: np.ndarray
    """Image array. `Array[uint, [H, W]|[H, W, 3]|[H, W, 4]]`"""

    mode: Literal['r', 'g', 'b', 'RGB', 'RGBA', 'gray']
    """Color mode"""

    @classmethod
    def load(cls, file: PathLike,
             alpha: bool = False) -> Self:
        """
        Load the image file

        :param file: Filepath of the image
        :param alpha: Keep alpha channel
        :return: ``ImageProcFactory``
        """

        img = cv2.imread(str(file))
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        mode = 'RGB'

        if alpha:
            img = cv2.cvtColor(img, cv2.COLOR_RGB2RGBA)
            img = np.array(img, dtype=np.uint8)
            mode = 'RGBA'

        return ImageProcFactory(img, mode)

    @property
    def height(self) -> int:
        """image height `H`"""
        return self.image.shape[0]

    @property
    def width(self) -> int:
        """image width `W`"""
        return self.image.shape[1]

    def select_channel(self, channel: IMAGE_CHANNEL_TYPE) -> Self:
        """Select `RGB` channel

        :param channel: {red', 'green', 'blue', 'r', 'g', 'b'}
        """

        if self.image.ndim < 3:
            raise TypeError('shape invalid for splitting channel')

        if channel in ('r', 'red'):
            img = cv2.split(self.image)[0]
            mode = 'r'
        elif channel in ('g', 'green'):
            img = cv2.split(self.image)[1]
            mode = 'g'
        elif channel in ('b', 'blue'):
            img = cv2.split(self.image)[2]
            mode = 'b'
        else:
            raise KeyError(f'{channel}')

        return attrs.evolve(self, image=img, mode=mode)

    def view_2d(self, flip: bool = True) -> Self:
        """view `RGB` or `RGBA` image as 2d array. `Array[uint, [H, W]]`"""
        if self.image.ndim == 2:
            img = self.image

        elif self.image.shape[2] == 4:
            w, h, _ = self.image.shape
            img = self.image.view(dtype=np.uint32).reshape((w, h))

        elif self.image.shape[2] == 3:
            r = self.image[:, :, 0]
            g = self.image[:, :, 1]
            b = self.image[:, :, 2]
            grayscale_image = 0.2989 * r + 0.5870 * g + 0.1140 * b
            img = grayscale_image.astype(np.uint8)

        else:
            raise RuntimeError(f'{self.image.shape} check')

        if flip:
            img = np.flipud(img)

        return attrs.evolve(self, image=img, mode='gray')

    # ============= #
    # Basic Process #
    # ============= #

    def cvt_gray(self) -> Self:
        """convert to grayscale"""
        if self.image.ndim == 2:
            img = np.uint8(self.image)
        elif self.image.shape[2] == 3:
            img = cv2.cvtColor(self.image, cv2.COLOR_BGR2GRAY)
        elif self.image.shape[2] == 4:
            img = cv2.cvtColor(self.image, cv2.COLOR_RGBA2GRAY)
        else:
            raise RuntimeError('')

        return attrs.evolve(self, image=img, mode='gray')

    def gaussian_blur(self, ksize: int, sigma: int) -> Self:
        img = cv2.GaussianBlur(self.image, (ksize, ksize), sigmaX=sigma, sigmaY=sigma)
        return attrs.evolve(self, image=img)

    def edge_detection(self,
                       lower_threshold: int = 30,
                       upper_threshold: int = 150) -> Self:
        grey_img = self.cvt_gray().image
        img = cv2.Canny(grey_img, lower_threshold, upper_threshold)
        return attrs.evolve(self, image=img)

    def binarize(self, threshold: int = 150) -> Self:
        _, img = cv2.threshold(self.image, threshold, 255, cv2.THRESH_BINARY)
        return attrs.evolve(self, image=img)

    def de_noise(self, h: int = 10, temp_win_size: int = 7, search_win_size: int = 21) -> Self:
        gray_img = self.cvt_gray().image
        dn = cv2.fastNlMeansDenoising(
            gray_img,
            h=h,
            templateWindowSize=temp_win_size,
            searchWindowSize=search_win_size
        )
        return attrs.evolve(self, image=dn)

    def local_maxima_image(self, channel: IMAGE_CHANNEL_TYPE, **kwargs) -> Self:
        """
        find the local maxima of the selection points.
        i.e., used in roi selection of the neuron before counting

        :param channel: color of image
        :return: `Array[uint, [H, W]]`
        """
        from skimage.morphology import local_maxima

        image = self.select_channel(channel).image
        if np.sum(image) == 0:
            return attrs.evolve(self, image=np.zeros_like(image, dtype=np.uint8))
        else:
            return attrs.evolve(self, image=local_maxima(image, **kwargs))


def recover_overexposure(img: np.ndarray,
                         alpha: float = 0.5,
                         beta: float = 0.5) -> np.ndarray:
    """recover saturated fluorescence image"""
    proc = cv2.addWeighted(img, alpha, np.zeros_like(img), beta, 0.0)
    return proc



from __future__ import annotations

from pathlib import Path
from typing import Union, Any, get_args, Final, Literal

import gspread
import numpy as np
import pandas as pd
import polars as pl
from typing_extensions import TypeAlias, Self

from neuralib.typing import PathLike, DataFrame
from neuralib.util.verbose import fprint

__all__ = [
    'SpreadSheetName',
    'WorkPageName',
    'DataIndex',
    #
    'GoogleSpreadSheet',
    'GoogleWorkSheet',
    #
    'upload_dataframe_to_spreadsheet'
]

SpreadSheetName: TypeAlias = str
"""spreadsheet name"""

WorkPageName: TypeAlias = str
"""workpage name of the spreadsheet"""

DataIndex: TypeAlias = Union[None, int, str, slice, list[int], list[str], np.ndarray]
"""data index type"""

VALUE_RENDER_OPT = Literal['FORMATTED_VALUE', 'UNFORMATTED_VALUE', 'FORMULA']
"""value render option for the cell"""


class GoogleWorkSheet:

    def __init__(self, worksheet: gspread.Worksheet,
                 primary_key: str | tuple[str, ...] = 'Data'):
        """

        :param worksheet: ``gspread.Worksheet``
        :param primary_key: Primary key of the worksheet.

            If str type, it must be one of the column name.

            If tuple str type, the primary key is join using "_" per row
        """
        self._worksheet = worksheet
        self._headers = tuple(self._worksheet.row_values(1))

        if isinstance(primary_key, str) and primary_key not in self._headers:
            raise ValueError(f'col not found: {primary_key}')
        elif isinstance(primary_key, (str, tuple)):
            self.primary_key: Final[str, tuple[str, ...]] = primary_key
        else:
            raise TypeError(f'{primary_key}')

    @classmethod
    def of(cls,
           name: SpreadSheetName,
           page: WorkPageName,
           service_account_path: PathLike,
           primary_key: str | tuple[str, ...] = 'Data') -> Self:
        """
        Get a worksheet from spreadsheet

        :param name: ``SpreadSheetName``
        :param page: ``WorkPageName``
        :param service_account_path: The path to the service account json file
        :param primary_key: Primary key of the worksheet.

            If str type, it must be one of the column name.

            If tuple str type, the primary key is join using "_" per row

        :return: ``GoogleWorkSheet``
        """
        sh = GoogleSpreadSheet(name, service_account_path, primary_key)

        if page not in sh:
            raise ValueError(f'{page} not found in spreadsheet: {name}')

        return sh.get_worksheet(page)

    @property
    def title(self) -> WorkPageName:
        """``WorkPageName``"""
        return self._worksheet.title

    @property
    def headers(self) -> list[str]:
        """list of worksheet header"""
        return list(self._headers)

    @property
    def primary_key_list(self) -> list[str]:
        """list of primary key of the worksheet"""
        primary = self.primary_key
        if isinstance(primary, str):
            return self.values(primary)
        elif isinstance(primary, tuple):
            ks = [self.values(p) for p in primary]

            if len(set(list(map(len, ks)))) != 1:
                print(set(list(map(len, ks))))
                raise RuntimeError(f'primary key cannot join properly due to different len in col: {self.primary_key}')

            return [
                '_'.join([str(it) for it in j])
                for j in (list(zip(*ks)))
            ]
        else:
            raise TypeError(f'{self.primary_key}')

    def get_range_value(self, a1_range_notation: str) -> list[Any]:
        """get values from range notation. i.e., `B1:S1` to get the list of content.

        If get values from 2D, the return order would be first column-wise, then row-wise

        """
        range_values = self._worksheet.range(a1_range_notation)
        return [it.value for it in range_values]

    def values(self, head: str) -> list[Any]:
        """get list of value from header"""
        col = self._col(head)
        return list(self._worksheet.col_values(col)[1:])

    def _row(self, data: DataIndex) -> Union[None, int, list[int], np.ndarray]:
        """
        Get row(s) index

        :param data: *str: first col value; *int: index
        :return:
        """
        if data is None:
            return None
        if isinstance(data, int):
            return data + 2  # skip header row + one-base
        elif isinstance(data, str):
            return self.primary_key_list.index(data) + 2
        elif isinstance(data, list):
            if len(data) == 0:
                return []
            return [self._row(it) for it in data]
        elif isinstance(data, (slice, np.ndarray)):
            return np.arange(len(self.primary_key_list))[data] + 2

        raise TypeError()

    def _col(self, head: str) -> int:
        """Get column index based on header (one-base)"""
        return self._headers.index(head) + 1

    def _rowcol(self, data: DataIndex, head: str) -> tuple[None | int | list[int], int]:
        """Get row in col indices"""
        return self._row(data), self._col(head)

    # noinspection PyTypeChecker
    def get_cell(self,
                 data: DataIndex,
                 head: str,
                 value_render_option: VALUE_RENDER_OPT = 'FORMATTED_VALUE'):
        """
        Get data from a cell

        :param data: ``DataIndex``
        :param head: header name
        :param value_render_option: ``VALUE_RENDER_OPT``: {'FORMATTED_VALUE', 'UNFORMATTED_VALUE', 'FORMULA'}
        :return:
        """
        if value_render_option not in get_args(VALUE_RENDER_OPT):
            raise ValueError('')

        row, col = self._rowcol(data, head)

        if row is None:
            return self._worksheet.col_values(col, value_render_option=value_render_option)[1:]
        elif isinstance(row, int):
            return self._worksheet.cell(row, col, value_render_option=value_render_option).value
        else:
            return [self._worksheet.cell(it, col, value_render_option=value_render_option).value for it in row]

    def update_cell(self, data: DataIndex,
                    head: str,
                    value: list[str] | str):
        """
        Update value in a cell

        :param data: ``DataIndex``
        :param head: header name
        :param value: value to be updated. str type if single field
        :return:
        """
        from gspread.utils import rowcol_to_a1

        row, col = self._rowcol(data, head)
        if row is None:
            if len(value) != len(self.primary_key_list):
                raise ValueError()

            data = []
            for it, v in enumerate(value):
                data.append({'range': rowcol_to_a1(it + 2, col), 'values': [[v]]})
                fprint(f'UPDATES: {rowcol_to_a1(it + 2, col)} from {self.get_cell(it, head)} -> {v}', vtype='io')

            self._worksheet.batch_update(data)

        elif isinstance(row, int):
            self._worksheet.update_cell(row, col, value)

        else:
            if len(value) != len(row):
                raise ValueError()

            data = []
            for it, v in zip(row, value):
                data.append({'range': rowcol_to_a1(it, col), 'values': [[v]]})
                fprint(f'UPDATES: {rowcol_to_a1(it, col)} from {self.get_cell(it - 2, head)}-> {v}', vtype='io')

            self._worksheet.batch_update(data)

    def clear(self):
        """Clears all cells in the worksheet"""
        self._worksheet.clear()

    def update(self, range_name, values=None, **kwargs):
        """Sets values in a cell range of the sheet"""
        self._worksheet.update(range_name, values, **kwargs)

    def to_pandas(self) -> pd.DataFrame:
        """Worksheet to pandas dataframe"""
        return pd.DataFrame(self._worksheet.get_all_records())

    def to_polars(self) -> pl.DataFrame:
        """Worksheet to polar dataframe"""
        return pl.DataFrame(self._worksheet.get_all_records(), nan_to_null=True)


# =========== #
# Spreadsheet #
# =========== #

class GoogleSpreadSheet:
    """
    Gspread module wrapper to access `google spreadsheet`

    .. seealso:: `<https://docs.gspread.org/en/latest/>`_


    .. code-block:: python

        service_account_path = ...  # The path to the service account json file
        sh = GoogleSpreadSheet('test_sheet', service_account_path)

        # get a worksheet from the spreadsheet
        ws = sh[WORK_SHEET_NAME]

    """

    def __init__(self, name: SpreadSheetName,
                 service_account_path: PathLike,
                 primary_key: str | tuple[str, ...] = 'Data'):
        """

        :param name: name of the spreadsheet
        :param service_account_path: The path to the service account json file
        :param primary_key: Primary key of the worksheet.

            If str type, it must be one of the column name.

            If tuple str type, the primary key is join using "_" per row
        """

        self._client = gspread.service_account(filename=service_account_path)

        self._sheet: gspread.Spreadsheet = self._client.open(name)
        self._worksheets: list[gspread.Worksheet] = self._sheet.worksheets()
        self._primary_key = primary_key

    def __len__(self):
        """number of worksheet"""
        return len(self._worksheets)

    def __contains__(self, item: WorkPageName) -> bool:
        return self.has_worksheet(item)

    def __getitem__(self, item: int | WorkPageName) -> GoogleWorkSheet:
        """get the worksheet.

        if int type, get worksheet from index.

        if str type, get worksheet from ``WorkPageName``

        """
        if isinstance(item, int):
            return GoogleWorkSheet(self._worksheets[item])
        elif isinstance(item, str):
            try:
                return self.get_worksheet(item)
            except ValueError:
                pass

            raise KeyError(f'{item} not found')
        else:
            raise TypeError(f'invalid item type: {type(item)}')

    @property
    def title(self) -> SpreadSheetName:
        """``SpreadSheetName``"""
        return self._sheet.title

    @property
    def worksheet_list(self) -> list[WorkPageName]:
        """list of ``WorkPageName``"""
        return [it.title for it in self._worksheets]

    def has_worksheet(self, page: WorkPageName) -> bool:
        """If has the worksheet, implement also in ``__contains__()``"""
        for w in self._worksheets:
            if w.title == page:
                return True
        return False

    def get_worksheet(self, page: WorkPageName) -> GoogleWorkSheet:
        """Get the worksheet. implement also in ``__get_item__()``"""
        for w in self._worksheets:
            if w.title == page:
                return GoogleWorkSheet(w, self._primary_key)

        raise ValueError(f'page not found: {page}')


# ========= #
# Utilities #
# ========= #

def upload_dataframe_to_spreadsheet(df: DataFrame,
                                    gspread_name: SpreadSheetName,
                                    worksheet_name: WorkPageName,
                                    service_account_path: Path | None,
                                    primary_key: str | tuple[str, ...] = 'Data') -> None:
    """
    Upload a dataframe to a gspread worksheet

    :param df: polars or pandas DataFrame
    :param gspread_name: spreadsheet name
    :param worksheet_name: worksheet name under the spreadsheet
    :param service_account_path: The path to the service account json file.
    :param primary_key: Primary key of the worksheet.
            If str type, it must be one of the column name.
            If tuple str type, the primary key is join using "_" per row
    """
    gs = GoogleSpreadSheet(gspread_name, service_account_path, primary_key)
    spreadsheet = gs._sheet

    if isinstance(df, pd.DataFrame):
        df = pl.from_pandas(df)

    if worksheet_name not in gs.worksheet_list:
        spreadsheet.add_worksheet(title=worksheet_name, rows=df.shape[0], cols=len(df.columns))
        fprint(f'ADD WORKSHEET: {worksheet_name}')
        gs = GoogleSpreadSheet(gspread_name, service_account_path, primary_key)  # refresh page

    worksheet = gs[worksheet_name]
    worksheet.clear()

    # cast dtype avoid serialization problem from gspread
    df = df.cast({pl.Datetime: pl.Utf8})

    data = [df.columns] + [field for field in df.iter_rows()]
    worksheet.update(data)



"""
Facemap
==========

This module provide ``Facemap`` result parsing
Including keypoints and pupil diameter data


Example of load the pupil area
--------------------------------

.. code-block:: python

    from neuralib.tracking.facemap import *

    directory = ...  # directory with output *proc.npy
    fmap = FaceMapResult.load(directory, track_type='pupil')
    area = fmap.get_pupil_area()



Example of launch the GUI with the existed results
--------------------------------------------------------

.. code-block:: python

    from neuralib.wrapper.facemap import *

    directory = ...  # directory with output *proc.npy (pathlike)
    with_keypoints = False  # if have keypoint tracking
    env_name = ...  # conda env name with facemap package installed (str)
    FaceMapResult.launch_facemap_gui(directory, with_keypoints=with_keypoints, env_name=env_name)



Example of load the keypoint result
------------------------------------

.. code-block:: python

    directory = ...
    fmap = FaceMapResult.load(directory, track_type='keypoints')

    plot_facemap_keypoints(fmap, frame_interval=(0, 100), keypoints=['eye(back)', 'eye(bottom)', 'eye(front)', 'eye(top)'])



.. toctree::
    :maxdepth: 1
    :caption: Notebook Demo

    ../notebooks/example_facemap

"""
from .core import *



from __future__ import annotations

import numpy as np
from matplotlib.axes import Axes

from neuralib.imaging.suite2p import Suite2PResult

__all__ = ['get_soma_pixel',
           'plot_soma_center']


def get_soma_pixel(s2p: Suite2PResult,
                   neuron_ids: np.ndarray | None = None,
                   *,
                   color_diff: bool = True,
                   include_overlap_pixel: bool = True) -> np.ndarray:
    """
    Get the image mask of the registered neuronal soma shape.
    The mask value equals to the neuron's id + 1.

    :param s2p: :class:`~neuralib.imaging.suite2p.core.Suite2PResult`
    :param neuron_ids: index ROIs array or bool mask array. `Array[int|bool, N]`
    :param color_diff: whether show color difference across neurons
    :param include_overlap_pixel: if taking overlap area into account
    :return: 2d imaging array, shape: (xpix, ypix)
    """
    neuron_pix = np.zeros((s2p.image_width, s2p.image_height))

    if neuron_ids is None:
        neuron_ids = np.arange(s2p.n_neurons)

    elif neuron_ids.dtype == bool:
        if len(neuron_ids) != s2p.stat.shape[0]:
            raise ValueError('invalid shape for the mask array')

        neuron_ids = np.nonzero(neuron_ids)[0]

    for i in neuron_ids:
        ypix = s2p.stat[i]['ypix']
        xpix = s2p.stat[i]['xpix']

        if not include_overlap_pixel:
            ypix = ypix[~s2p.stat[i]['overlap']]
            xpix = xpix[~s2p.stat[i]['overlap']]

        neuron_pix[xpix, ypix] = i + 1 if color_diff else 1

    return neuron_pix


def plot_soma_center(ax: Axes,
                     s2p: Suite2PResult,
                     neuron_ids: np.ndarray | None = None,
                     *,
                     invert_xy: bool = True,
                     with_index: bool = True,
                     font_size: float = 5,
                     **kwargs) -> None:
    """Plot center of roi soma and its corresponding index

    :param ax: :class:`matplotlib.axes.Axes`
    :param s2p: :class:`~neuralib.imaging.suite2p.core.Suite2PResult`
    :param neuron_ids: index ROIs array or bool mask array. `Array[int|bool, N]`
    :param invert_xy: If invert the FOV xy pixel
    :param with_index: whether plot the index ROI index nearby
    :param font_size: font size of the text
    :param kwargs: additional args to ``plt.scatter()``
    """
    if neuron_ids is None:
        n_neuron = s2p.stat.shape[0]
        neuron_ids = np.arange(n_neuron)

    elif neuron_ids.dtype == bool:
        if len(neuron_ids) != s2p.stat.shape[0]:
            raise ValueError('invalid shape for the mask array')
        neuron_ids = np.nonzero(neuron_ids)[0]

    coords = np.array([(np.mean(s2p.stat[i]['ypix' if invert_xy else 'xpix']),
                        np.mean(s2p.stat[i]['xpix' if invert_xy else 'ypix'])) for i in neuron_ids])

    ax.scatter(coords[:, 0], coords[:, 1], **kwargs)

    if with_index:
        for (x, y), idx in zip(coords, neuron_ids):
            ax.text(x, y, str(idx), fontsize=font_size)



"""
Value dispatch function
=======================

:author:
    Ta-Shun Su

Work with AbstractParser.

>>> from neuralib.argp import AbstractParser
>>> class Test(AbstractParser, DispatchOption):
...     target: str = DispatchOption.argument(
...         '--run'
...     )
...
...     @classmethod
...     def EPILOG(cls): # build parser epilog
...         return f'''\
... Command (--run)
... {textwrap.indent(Test.parser_command_epilog(), "  ")}
... '''
...
...     def run(self):
...         # dispaatch function call according to self.target
...         self.invoke_command(self.target)
...
...     @DispatchOption.dispatch('A')
...     def run_a(self):
...         '''doc A.'''
...         print('a')
...     @DispatchOption.dispatch('B')
...     def run_b(self):
...         '''doc B'''
...         print('b')
"""

from typing import Union, NamedTuple, Optional, Callable, TypeVar, Any, Final, Type

from .core import Argument

__all__ = [
    'dispatch',
    'list_commands',
    'find_command',
    'invoke_command',
    'invoke_group_command',
    'DispatchOption',
    'DispatchCommandNotFound'
]

T = TypeVar('T')
missing = object()

NEURALIB_DISPATCH_INFO = '__NEURALIB_DISPATCH_INFO__'


def dispatch(command: str,
             alias: Union[str, list[str]] = None,
             group: str = None):
    """A decorator that mark a function a dispatch target function.

    All functions decorated in same dispatch group should have save
    function signature (at least for non-default parameters). For example:

    >>> @dispatch('A')
    ... def function_a(self, a, b, c=None):
    ...     pass
    ... @dispatch('B')
    ... def function_b(self, a, b, d=None):
    ...     pass
    ... def run_function(self):
    ...     invoke_command(self, 'A', a, b)


    """
    if len(command) == 0:
        raise ValueError('empty command string')

    if alias is None:
        alias = []
    elif isinstance(alias, str):
        alias = [alias]

    def _dispatch(f):
        if hasattr(f, NEURALIB_DISPATCH_INFO):
            raise RuntimeError()

        setattr(f, NEURALIB_DISPATCH_INFO, DispatchCommand(command, tuple(alias), group, f))
        return f

    return _dispatch


class DispatchCommand(NamedTuple):
    command: str
    aliases: tuple[str, ...]
    group: Optional[str]
    func: Callable

    def __call__(self, host: T, *args, **kwargs) -> Any:
        return self.func(host, *args, **kwargs)

    @property
    def doc(self) -> Optional[str]:
        return self.func.__doc__


class DispatchCommandNotFound(RuntimeError):
    def __init__(self, command: str, group: Optional[str] = None):
        if group is None:
            message = f'command {command} not found'
        else:
            message = f'command {group}:{command} not found'

        super().__init__(message)


def list_commands(host: Union[T, Type[T]], group: Optional[str] = missing) -> list[DispatchCommand]:
    """list all dispatch-decoratored function info in *host*.

    :param host:
    :param group: dispatch group.
    :return: list of DispatchCommand
    """
    ret = []

    info: DispatchCommand

    if isinstance(host, type):
        host_type = host
    else:
        host_type = type(host)

    for attr in dir(host_type):
        attr_value = getattr(host_type, attr)
        if (info := getattr(attr_value, NEURALIB_DISPATCH_INFO, None)) is not None:
            if group is missing or group == info.group:
                ret.append(info)

    return ret


def find_command(host: T, command: str, group: Optional[str] = missing) -> Optional[DispatchCommand]:
    """find dispatch-decoratored function in *host* according to *command*.

    :param host:
    :param command: command or command alias
    :param group: dispatch group
    :return: found DispatchCommand
    """
    info: DispatchCommand

    host_type = type(host)
    for attr in dir(host_type):
        attr_value = getattr(host_type, attr)
        if (info := getattr(attr_value, NEURALIB_DISPATCH_INFO, None)) is not None:
            if group is missing or group == info.group:
                if command == info.command or command in info.aliases:
                    return info

    return None


def invoke_command(host: T, command: str, *args, **kwargs) -> Any:
    """invoke a dispatch-decoratored function in default group.

    :param host:
    :param command: command or command alias
    :param args: dispatch-decoratored function positional arguments
    :param kwargs: dispatch-decoratored function keyword arguments
    :return: function return
    :raise DispatchCommandNotFound:
    """
    if (info := find_command(host, command, None)) is None:
        raise DispatchCommandNotFound(command)
    return info(host, *args, **kwargs)


def invoke_group_command(host: T, group: str, command: str, *args, **kwargs) -> Any:
    """invoke a dispatch-decoratored function in certain group.

    :param host:
    :param group: dispatch group
    :param command: command or command alias
    :param args: dispatch-decoratored function positional arguments
    :param kwargs: dispatch-decoratored function keyword arguments
    :return: function return
    :raise DispatchCommandNotFound:
    """
    if (info := find_command(host, command, group)) is None:
        raise DispatchCommandNotFound(command, group)
    return info(host, *args, **kwargs)


class DispatchArgument(Argument):
    pass


class DispatchOption:
    argument: Final = DispatchArgument

    def list_commands(self, group: Optional[str] = missing) -> list[DispatchCommand]:
        """list all dispatch-decoratored function info in *host*.

        :param group: dispatch group.
        :return: list of DispatchCommand
        """
        return list_commands(self, group)

    def find_command(self, command: str, group: Optional[str] = missing) -> Optional[DispatchCommand]:
        """find dispatch-decoratored function in *host* according to *command*.

        :param command: command or command alias
        :param group: dispatch group
        :return: found DispatchCommand
        """
        return find_command(self, command, group)

    def invoke_command(self, command: str, *args, **kwargs) -> Any:
        """invoke a dispatch-decoratored function in default group.

        :param command: command or command alias
        :param args: dispatch-decoratored function positional arguments
        :param kwargs: dispatch-decoratored function keyword arguments
        :return: function return
        :raise DispatchCommandNotFound:
        """
        return invoke_command(self, command, *args, **kwargs)

    def invoke_group_command(self, group: str, command: str, *args, **kwargs) -> Any:
        """invoke a dispatch-decoratored function in certain group.

        :param group: dispatch group
        :param command: command or command alias
        :param args: dispatch-decoratored function positional arguments
        :param kwargs: dispatch-decoratored function keyword arguments
        :return: function return
        :raise DispatchCommandNotFound:
        """
        return invoke_group_command(self, group, command, *args, **kwargs)

    @classmethod
    def dispatch(cls, command: str,
                 alias: Union[str, list[str]] = None,
                 group: str = None):
        return dispatch(command, alias, group)

    @classmethod
    def parser_command_epilog(cls, group: Optional[str] = missing) -> str:
        from rich.console import Console
        from rich.table import Table
        from rich import box
        console = Console()

        with console.capture() as capture:
            table = Table(show_header=False, show_edge=False, box=box.SIMPLE)
            table.add_column('command', min_width=18)
            table.add_column('desp')

            for info in list_commands(cls, group):
                header = info.command
                if len(info.aliases) > 0:
                    header += ' (' + ', '.join(info.aliases) + ')'

                content = info.doc or ''
                content = content.split('\n')[0].strip()

                table.add_row(header, content)

            console.print(table)

        return capture.get()



from __future__ import annotations

import pickle
from pathlib import Path
from typing import TypedDict, NamedTuple

import h5py
import numpy as np
import polars as pl
from typing_extensions import Self

from neuralib.typing import PathLike

__all__ = [
    'Joint',
    'DeepLabCutMeta',
    'DeepLabCutModelConfig',
    'DeepLabCutResult',
    'load_dlc_result'
]

Joint = str


class DeepLabCutMeta(TypedDict):
    start: float
    stop: float
    run_duration: float
    Scorer: str
    model_config: DeepLabCutModelConfig
    fps: float
    batch_size: int
    frame_dimensions: tuple[int, int]
    nframes: int
    iteration: int
    training_set_fraction: float
    cropping: bool
    cropping_parameters: list[tuple[float, float, float, float]]


class DeepLabCutModelConfig(TypedDict):
    stride: float
    weigh_part_predictions: bool
    weigh_negatives: bool
    fg_fraction: float
    mean_pixel: list[float]
    shuffle: bool
    snapshot_prefix: str
    log_dir: str
    global_scale: float
    location_refinement: bool
    locref_stdev: float
    locref_loss_weight: float
    locref_huber_loss: bool
    optimizer: str
    intermediate_supervision: bool
    intermediate_supervision_layer: int
    regularize: bool
    weight_decay: float
    crop_pad: int
    scoremap_dir: str
    batch_size: int
    dataset_type: str
    deterministic: bool
    mirror: bool
    pairwise_huber_loss: bool
    weigh_only_present_joints: bool
    partaffinityfield_predict: bool
    pairwise_predict: bool
    all_joints: list[list[int]]
    all_joints_names: list[Joint]
    dataset: str
    init_weights: str
    net_type: str
    num_joints: int
    num_outputs: int


class DeepLabCutJoint(NamedTuple):
    source: DeepLabCutResult
    name: Joint
    dat: pl.DataFrame

    def with_lh_filter(self, lh: float) -> Self:
        expr = pl.col('likelihood') >= lh
        dat = self.dat.with_columns(
            pl.when(expr).then(pl.col('x')).otherwise(np.nan),
            pl.when(expr).then(pl.col('y')).otherwise(np.nan)
        )
        return self._replace(dat=dat)

    @property
    def x(self) -> np.ndarray:
        return self.dat['x'].to_numpy()

    @property
    def y(self) -> np.ndarray:
        return self.dat['y'].to_numpy()

    @property
    def xy(self) -> np.ndarray:
        return self.dat['x', 'y'].to_numpy()

    @property
    def t(self) -> np.ndarray:
        return self.source.time

    @property
    def lh(self) -> np.ndarray:
        return self.dat['likelihood'].to_numpy()


class DeepLabCutResult:

    def __init__(self,
                 dat: pl.DataFrame,
                 meta: DeepLabCutMeta,
                 filtered: bool,
                 time: np.ndarray | None = None):
        """

        :param dat: Deeplabcut results as polars dataframe
        :param meta: Deeplabcut meta typeddict
        :param filtered: If the Deeplabcut results is filtered or not
        :param time: 1D time array for each tracked frames. If None, then assume stable DAQ and calculated from meta.
        """
        self.dat = dat
        self._meta = meta
        self._filtered = filtered

        self._time = time if time is not None else self._default_time()

    def __getitem__(self, item: Joint) -> DeepLabCutJoint:
        """Get data from a specific joint"""
        cols = ('x', 'y', 'likelihood')
        dat = self.dat[[f'{item}_{col}' for col in cols]]
        dat.columns = cols
        return DeepLabCutJoint(self, item, dat)

    @property
    def is_filtered(self) -> bool:
        return self._filtered

    @property
    def meta(self) -> DeepLabCutMeta:
        return self._meta

    @property
    def joints(self) -> list[Joint]:
        """list of labelled joints"""
        return self.meta['model_config']['all_joints_names']

    @property
    def fps(self) -> float:
        return self.meta['fps']

    @property
    def nframes(self) -> int:
        return self.meta['nframes']

    @property
    def time(self) -> np.ndarray:
        if len(self._time) != self.dat.shape[0]:
            raise ValueError('time array has wrong shape, mismatch with dat')
        return self._time

    def _default_time(self):
        """Based on meta file, assume stable frame acquisition"""
        return np.linspace(0, self.nframes / self.fps, self.nframes)

    def with_global_lh_filter(self, lh: float) -> Self:
        """
        With global likelihood filter
        :param lh: likelihood threshold
        :return: ``DeepLabCutResult``
        """
        for j in self.joints:
            self.dat = (
                self.dat
                .with_columns((pl.col(f'{j}_likelihood') >= lh).alias('valid'))
                .with_columns(pl.when(pl.col('valid')).then(pl.col(f'{j}_x')).otherwise(np.nan))
                .with_columns(pl.when(pl.col('valid')).then(pl.col(f'{j}_y')).otherwise(np.nan))
            )

        return self


def load_dlc_result(file: PathLike,
                    meta_file: PathLike,
                    time: np.ndarray | None = None) -> DeepLabCutResult:
    """
    Load DeepLabCut result from file

    :param file: DeepLabCut result filepath. supports both `.h5` and `.csv`
    :param meta_file: DeepLabCut meta filepath. should be the `.pickle`. TODO Cannot it be inferred according to file?
    :param time: time array for each sample point. If None, then assume stable DAQ for using total frames and fps info in meta
    :return: ``DeepLabCutResult``
    """
    file = Path(file)
    meta_file = Path(meta_file)
    meta = _load_meta(meta_file)

    if file.suffix in ('.h5', '.hdf5'):
        try:
            import pytables  # noqa: F401
            import pandas  # noqa: F401
            df = _load_dlc_h5_table(file)
        except ImportError:
            df = _load_dlc_h5(file, meta)
    elif file.suffix == '.csv':
        df = _load_dlc_csv(file)
    else:
        raise ValueError(f'Unsupported file type: {file.suffix}')

    return DeepLabCutResult(df, meta, filtered=('filtered' in file.name), time=time)


def _load_dlc_h5_table(file) -> pl.DataFrame:
    import pandas as pd

    df = pd.read_hdf(file)

    scorers = list(df.columns.levels[0])
    bodyparts = list(df.columns.levels[1])
    coords = list(df.columns.levels[2])

    assert len(scorers) == 1
    scorer = scorers[0]

    data = {
        f'{b}_{c}': df[(scorer, b, c)]
        for b in bodyparts
        for c in coords
    }

    ret = pl.DataFrame(data)

    return ret


def _load_dlc_h5(file, meta) -> pl.DataFrame:
    dat = h5py.File(file)['df_with_missing']['table']['values_block_0']
    return pl.from_numpy(dat, schema=[
        f'{joint}_{it}'
        for joint in meta['model_config']['all_joints_names']
        for it in ('x', 'y', 'likelihood')
    ])


def _load_dlc_csv(file) -> pl.DataFrame:
    cols = ['']
    with file.open() as f:
        f.readline()  # skip first line
        parts = f.readline().strip().split(',')[1::3]
        cols.extend([f'{p}_{it}' for p in parts for it in ('x', 'y', 'likelihood')])

    df = pl.read_csv(file, skip_rows=3, has_header=False, new_columns=cols)[:, 1:]

    return df


def _load_meta(meta_file) -> DeepLabCutMeta:
    if meta_file.suffix not in ('.pkl', '.pickle'):
        raise ValueError(f'{meta_file} is not a pickle file')

    # meta
    with meta_file.open('rb') as f:
        meta = pickle.load(f)['data']

    # copy to typeddict
    meta['model_config'] = meta['DLC-model-config file']
    meta['iteration'] = meta['iteration (active-learning)']
    meta['training_set_fraction'] = meta['training set fraction']

    return meta
